# -*- coding: utf-8 -*-

__all__ = [
    'get_coverage_level_fm_terms',
    'get_fm_terms_by_level_as_list',
    'get_layer_calcrule_id', 
    'get_layer_level_fm_terms',
    'get_policytc_ids',
    'get_sub_layer_calcrule_id',
    'get_sub_layer_non_coverage_level_fm_terms',
    'unified_canonical_fm_profile_by_level',
    'unified_canonical_fm_profile_by_level_and_term_group'
]

import io
import itertools
import json
import six

import pandas as pd

from .exceptions import OasisException
from .metadata import OASIS_FM_LEVELS


def get_coverage_level_fm_terms(level_unified_canonical_profile, level_fm_agg_profile, level_fm_items, canexp_df, canacc_df):

    lid = level_fm_items[0]['level_id']

    if lid != OASIS_FM_LEVELS['coverage']['id']:
        raise OasisException('Invalid FM level ID {} for generating coverage level FM terms - expected to be {}'.format(lid, OASIS_FM_LEVELS['coverage']['id']))

    lufcp = level_unified_canonical_profile

    lfmap = level_fm_agg_profile

    agg_key = tuple(v['field'].lower() for v in six.itervalues(lfmap['FMAggKey']))

    li = sorted([it for it in six.itervalues(level_fm_items)], key=lambda it: tuple(it[k] for k in agg_key))

    can_df = pd.merge(canexp_df, canacc_df, left_on='accntnum', right_on='accntnum')

    get_can_item = lambda canexp_id, canacc_id, policy_num: can_df[(can_df['row_id_x']==canexp_id+1) & (can_df['row_id_y']==canacc_id+1) & (can_df['policynum']==policy_num)].iloc[0]

    for it, i in itertools.chain((it, i) for i, (key, group) in enumerate(itertools.groupby(li, key=lambda it: tuple(it[k] for k in agg_key))) for it in group):
        it['agg_id'] = i + 1

        can_item = get_can_item(it['canexp_id'], it['canacc_id'], it['policy_num'])

        can_item_ded = can_item.get(it['ded_elm']) or 0.0
        it['deductible'] = (can_item_ded if can_item_ded >= 1 else it['tiv']*can_item_ded) or 0.0

        can_item_ded_min = can_item.get(it['ded_min_elm']) or 0.0
        it['deductible_min'] = (can_item_ded_min if can_item_ded_min >= 1 else it['tiv']*can_item_ded_min) or 0.0

        can_item_ded_max = can_item.get(it['ded_max_elm']) or 0.0
        it['deductible_max'] = (can_item_ded_max if can_item_ded_max >= 1 else it['tiv']*can_item_ded_max) or 0.0

        can_item_lim = can_item.get(it['lim_elm']) or 0.0
        it['limit'] = (can_item_lim if can_item_lim >= 1 else it['tiv']*can_item_lim) or 0.0

        it['share'] = can_item.get(it['shr_elm']) or 0.0

        it['calcrule_id'] = get_sub_layer_calcrule_id(it['deductible'], it['deductible_min'], it['deductible_max'], it['limit'])

        yield it


def get_layer_level_fm_terms(level_unified_canonical_profile, level_fm_agg_profile, level_fm_items, canexp_df, canacc_df):

    lid = level_fm_items[0]['level_id']

    if lid != OASIS_FM_LEVELS['layer']['id']:
        raise OasisException('Invalid FM level ID {} for generating coverage level FM terms - expected to be {}'.format(lid, OASIS_FM_LEVELS['layer']['id']))

    lufcp = level_unified_canonical_profile

    lfmap = level_fm_agg_profile

    agg_key = tuple(v['field'].lower() for v in six.itervalues(lfmap['FMAggKey']))

    li = sorted([it for it in six.itervalues(level_fm_items)], key=lambda it: tuple(it[k] for k in agg_key))

    can_df = pd.merge(canexp_df, canacc_df, left_on='accntnum', right_on='accntnum')

    get_can_item = lambda canexp_id, canacc_id, policy_num: can_df[(can_df['row_id_x']==canexp_id+1) & (can_df['row_id_y']==canacc_id+1) & (can_df['policynum']==policy_num)].iloc[0]

    ded_fld = lufcp[1].get('deductible')
    ded_elm = ded_fld['ProfileElementName'].lower() if ded_fld else None

    lim_fld = lufcp[1].get('limit')
    lim_elm = lim_fld['ProfileElementName'].lower() if lim_fld else None

    shr_fld = lufcp[1].get('share')
    shr_elm = shr_fld['ProfileElementName'].lower() if shr_fld else None

    for it, i in itertools.chain((it, i) for i, (key, group) in enumerate(itertools.groupby(li, key=lambda it: tuple(it[k] for k in agg_key))) for it in group):
        it['agg_id'] = i + 1

        can_item = get_can_item(it['canexp_id'], it['canacc_id'], it['policy_num'])

        it['ded_elm'] = ded_elm
        it['deductible'] = can_item.get(ded_elm) or 0.0
        it['attachment'] = it['deductible']
        it['deductible_min'] = it['deductible_max'] = 0.0

        it['lim_elm'] = lim_elm
        it['limit'] = can_item.get(lim_elm) or 0.0

        it['shr_elm'] = shr_elm
        it['share'] = can_item.get(shr_elm) or 0.0

        it['calcrule_id'] = get_layer_calcrule_id(it['attachment'], it['limit'], it['share'])

        yield it


def get_sub_layer_non_coverage_level_fm_terms(level_unified_canonical_profile, level_fm_agg_profile, level_fm_items, canexp_df, canacc_df):

    lid = level_fm_items[0]['level_id']

    r = range(OASIS_FM_LEVELS['combined']['id'], OASIS_FM_LEVELS['layer']['id'])
    if lid not in r:
        raise OasisException('Invalid FM level ID {} for generating sub-layer non-coverage level FM terms - expected to be in the range {}...{}'.format(lid, r.start, r.stop - 1))

    lufcp = level_unified_canonical_profile

    lfmap = level_fm_agg_profile

    agg_key = tuple(v['field'].lower() for v in six.itervalues(lfmap['FMAggKey']))

    li = sorted([it for it in six.itervalues(level_fm_items)], key=lambda it: tuple(it[k] for k in agg_key))

    can_df = pd.merge(canexp_df, canacc_df, left_on='accntnum', right_on='accntnum')

    get_can_item = lambda canexp_id, canacc_id, policy_num: can_df[(can_df['row_id_x']==canexp_id+1) & (can_df['row_id_y']==canacc_id+1) & (can_df['policynum']==policy_num)].iloc[0]

    ded_fld = lufcp[1].get('deductible')
    ded_elm = ded_fld['ProfileElementName'].lower() if ded_fld else None

    ded_min_fld = lufcp[1].get('deductiblemin')
    ded_min_elm = ded_min_fld['ProfileElementName'].lower() if ded_min_fld else None

    ded_max_fld = lufcp[1].get('deductiblemax')
    ded_max_elm = ded_max_fld['ProfileElementName'].lower() if ded_max_fld else None

    lim_fld = lufcp[1].get('limit')
    lim_elm = lim_fld['ProfileElementName'].lower() if lim_fld else None

    shr_fld = lufcp[1].get('share')
    shr_elm = shr_fld['ProfileElementName'].lower() if shr_fld else None

    for it, i in itertools.chain((it, i) for i, (key, group) in enumerate(itertools.groupby(li, key=lambda it: tuple(it[k] for k in agg_key))) for it in group):
        it['agg_id'] = i + 1

        can_item = get_can_item(it['canexp_id'], it['canacc_id'], it['policy_num'])

        it['ded_elm'] = ded_elm
        can_item_ded = can_item.get(ded_elm) or 0.0
        it['deductible'] = (can_item_ded if can_item_ded >= 1 else it['tiv']*can_item_ded) or 0.0

        it['ded_min_elm'] = ded_min_elm
        can_item_ded_min = can_item.get(ded_min_elm) or 0.0
        it['deductible_min'] = (can_item_ded_min if can_item_ded_min >= 1 else it['tiv']*can_item_ded_min) or 0.0

        it['ded_max_elm'] = ded_max_elm
        can_item_ded_max = can_item.get(ded_max_elm) or 0.0
        it['deductible_max'] = (can_item_ded_max if can_item_ded_max >= 1 else it['tiv']*can_item_ded_max) or 0.0

        it['lim_elm'] = lim_elm
        can_item_lim = can_item.get(lim_elm) or 0.0
        it['limit'] = (can_item_lim if can_item_lim >= 1 else it['tiv']*can_item_lim) or 0.0

        it['calcrule_id'] = get_sub_layer_calcrule_id(it['deductible'], it['deductible_min'], it['deductible_max'], it['limit'])

        yield it


def get_fm_terms_by_level_as_list(level_unified_canonical_profile, level_fm_agg_profile, level_fm_items, canexp_df, canacc_df):

    level_id = level_fm_items[0]['level_id']

    if level_id == OASIS_FM_LEVELS['coverage']['id']:
        return [it for it in get_coverage_level_fm_terms(level_unified_canonical_profile, level_fm_agg_profile, level_fm_items, canexp_df, canacc_df)]
    elif level_id in range(OASIS_FM_LEVELS['combined']['id'], OASIS_FM_LEVELS['layer']['id']):
        return [it for it in get_sub_layer_non_coverage_level_fm_terms(level_unified_canonical_profile, level_fm_agg_profile, level_fm_items, canexp_df, canacc_df)]
    elif level_id == OASIS_FM_LEVELS['layer']['id']:
        return [it for it in get_layer_level_fm_terms(level_unified_canonical_profile, level_fm_agg_profile, level_fm_items, canexp_df, canacc_df)]

def get_policytc_ids(fm_items_df):

    columns = [
        col for col in fm_items_df.columns if not col in ('limit', 'deductible', 'deductible_min', 'deductible_max', 'attachment', 'share', 'calcrule_id',)
    ]

    policytc_df = fm_items_df.drop(columns, axis=1).drop_duplicates()
    
    for col in policytc_df.columns:
        policytc_df[col] = policytc_df[col].astype(float) if col != 'calcrule_id' else policytc_df[col].astype(int)

    policytc_df['index'] = range(1, len(policytc_df) + 1)

    policytc_ids = {
        i: {
            'limit': policytc_df.iloc[i - 1]['limit'],
            'deductible': policytc_df.iloc[i - 1]['deductible'],
            'deductible_min': policytc_df.iloc[i - 1]['deductible_min'],
            'deductible_max': policytc_df.iloc[i - 1]['deductible_max'],
            'attachment': policytc_df.iloc[i - 1]['attachment'],
            'share': policytc_df.iloc[i - 1]['share'],
            'calcrule_id': int(policytc_df.iloc[i - 1]['calcrule_id'])
        } for i in policytc_df['index']
    }

    return policytc_ids


def get_layer_calcrule_id(att=0, lim=9999999999, shr=1):

    if att > 0 or lim > 0 or shr > 0:
        return 2


def get_sub_layer_calcrule_id(ded, ded_min, ded_max, lim, ded_code=0, lim_code=0):

    if ded == ded_code == ded_min == ded_max == lim == lim_code == 0:
        return 12
    elif (ded > 0 and ded_code == 0) and (ded_min == ded_max == 0) and (lim > 0 and lim_code == 0):
        return 1
    elif (ded > 0 and ded_code == 2) and (ded_min == ded_max == 0) and (lim > 0 and lim_code == 0):
        return 4
    elif (ded > 0 and ded_code == 1) and (ded_min == ded_max == 0) and (lim > 0 and lim_code == 1):
        return 5
    elif (ded > 0 and ded_code == 2) and (ded_min == ded_max == 0) and (lim == lim_code == 0):
        return 6
    elif (ded == ded_code == 0) and (ded_min == 0 and ded_max > 0) and (lim > 0 and lim_code == 0):
        return 7
    elif (ded == ded_code == 0) and (ded_min > 0 and ded_max == 0) and (lim > 0 and lim_code == 0):
        return 8
    elif (ded == ded_code == 0) and (ded_min == 0 and ded_max > 0) and (lim == lim_code == 0):
        return 10
    elif (ded == ded_code == 0) and (ded_min > 0 and ded_max == 0) and (lim == lim_code == 0):
        return 11
    elif (ded > 0 and ded_code == 0) and (ded_min == ded_max == 0) and (lim == lim_code == 0):
        return 12
    elif (ded == ded_code == 0) and (ded_min > 0 and ded_max > 0) and (lim == lim_code == 0):
        return 13
    elif (ded == ded_code == 0) and (ded_min == ded_max == 0) and (lim > 0 and lim_code == 0):
        return 14
    elif (ded == ded_code == 0) and (ded_min == ded_max == 0) and (lim > 0 and lim_code == 1):
        return 15
    elif (ded > 0 and ded_code == 1) and (ded_min == ded_max == 0) and (lim == lim_code == 0):
        return 16
    elif (ded > 0 and ded_code == 1) and (ded_min > 0 and ded_max > 0) and (lim == lim_code == 0):
        return 19
    elif (ded > 0 and ded_code == 2) and (ded_min > 0 and ded_max > 0) and (lim == lim_code == 0):
        return 21


def unified_canonical_fm_profile_by_level(profiles=[], profile_paths=[]):

    if not (profiles or profile_paths):
        raise OasisException('A list of canonical profiles (loc. or acc.) or a list of canonical profiles paths must be provided')

    if not profiles:
        for pp in profile_paths:
            with io.open(pp, 'r', encoding='utf-8') as f:
                profiles.append(json.load(f))

    comb_prof = {k:v for p in profiles for k, v in ((k, v) for k, v in six.iteritems(p) if 'FMLevel' in v)}
    
    return {
        int(k):{v['ProfileElementName']:v for v in g} for k, g in itertools.groupby(sorted(six.itervalues(comb_prof), key=lambda v: v['FMLevel']), key=lambda v: v['FMLevel'])
    }


def unified_canonical_fm_profile_by_level_and_term_group(profiles=[], profile_paths=[]):

    if not (profiles or profile_paths):
        raise OasisException('A list of canonical profiles (loc. or acc.) or a list of canonical profiles paths must be provided')

    comb_prof = unified_canonical_fm_profile_by_level(profiles=profiles, profile_paths=profile_paths)

    return {
        k:{
            _k:{v['FMTermType'].lower():v for v in g} for _k, g in itertools.groupby(sorted(six.itervalues(comb_prof[k]), key=lambda v: v['FMTermGroupID']), key=lambda v: v['FMTermGroupID'])
        } for k in comb_prof
    }



# -*- coding: utf-8 -*-

from itertools import chain

__all__ = [
    'OasisFilesPipeline'
]


class OasisFilesPipeline(object):
    def __init__(
        self,
        model_key=None,
        source_exposures_file_path=None,
        source_accounts_file_path=None,
        canonical_exposures_file_path=None,
        canonical_accounts_file_path=None,
        model_exposures_file_path=None,
        keys_file_path=None,
        keys_errors_file_path=None
    ):
        self._model_key = model_key
        self._source_exposures_file_path = source_exposures_file_path
        self._source_accounts_file_path = source_accounts_file_path
        self._canonical_exposures_file_path = canonical_exposures_file_path
        self._canonical_accounts_file_path = canonical_accounts_file_path
        self._model_exposures_file_path = model_exposures_file_path
        self._keys_file_path = keys_file_path
        self._keys_errors_file_path = keys_errors_file_path

        self._items_file_path = None
        self._coverages_file_path = None
        self._gulsummaryxref_file_path = None

        self._fm_policytc_file_path = None
        self._fm_profile_file_path = None
        self._fm_programme_file_path = None
        self._fm_xref_file_path = None
        self._fmsummaryxref_file_path = None

        self._source_files = {
            'source_exposures': self._source_exposures_file_path,
            'source_accounts': self._source_accounts_file_path 
        }

        self._intermediate_files = {
            'canonical_exposures': self._canonical_exposures_file_path,
            'canonical_accounts': self._canonical_accounts_file_path,
            'model_exposures': self._model_exposures_file_path,
            'keys': self._keys_file_path,
            'keys_errors': self._keys_errors_file_path
        }

        self._gul_files = {
            'items': self._items_file_path,
            'coverages': self._coverages_file_path,
            'gulsummaryxref': self._gulsummaryxref_file_path
        }

        self._fm_files = {
            'fm_policytc': self._fm_policytc_file_path,
            'fm_profile': self._fm_profile_file_path,
            'fm_programme': self._fm_programme_file_path,
            'fm_xref': self._fm_xref_file_path,
            'fmsummaryxref': self._fmsummaryxref_file_path
        }

        self._oasis_files = {k:v for k, v in chain(self._gul_files.items(), self._fm_files.items())}

    def __str__(self):
        return '{}: {}'.format(self.__repr__(), self.model_key)

    def __repr__(self):
        return '{}: {}'.format(self.__class__, self.__dict__)

    def _repr_pretty_(self, p, cycle):
        p.text(str(self) if not cycle else '...')

    @property
    def model_key(self):
        """
        Model key property - getter only.

            :getter: Gets the key of model to which the pipeline is attached.
        """
        return self._model_key

    @property
    def source_exposures_file_path(self):
        """
        Source exposures file path property.

            :getter: Gets the file path
            :setter: Sets the current file path to the specified file path
        """
        return self._source_exposures_file_path

    @source_exposures_file_path.setter
    def source_exposures_file_path(self, p):
        self._source_exposures_file_path = self.source_files['source_exposures'] = p

    @property
    def source_accounts_file_path(self):
        """
        Source accounts file path property.

            :getter: Gets the file path
            :setter: Sets the current file path to the specified file path
        """
        return self._source_accounts_file_path

    @source_accounts_file_path.setter
    def source_accounts_file_path(self, p):
        self._source_accounts_file_path = self.source_files['source_accounts'] = p

    @property
    def canonical_exposures_file_path(self):
        """
        Canonical exposures file path property.

            :getter: Gets the file path
            :setter: Sets the current file path to the specified file path
        """
        return self._canonical_exposures_file_path

    @canonical_exposures_file_path.setter
    def canonical_exposures_file_path(self, p):
        self._canonical_exposures_file_path = self.intermediate_files['canonical_exposures'] = p

    @property
    def canonical_accounts_file_path(self):
        """
        Canonical accounts file path property.

            :getter: Gets the file path
            :setter: Sets the current file path to the specified file path
        """
        return self._canonical_accounts_file_path

    @canonical_accounts_file_path.setter
    def canonical_accounts_file_path(self, p):
        self._canonical_accounts_file_path = self.intermediate_files['canonical_accounts'] = p

    @property
    def model_exposures_file_path(self):
        """
        Model exposures file path property.

            :getter: Gets the file path
            :setter: Sets the current file path to the specified file path
        """
        return self._model_exposures_file_path

    @model_exposures_file_path.setter
    def model_exposures_file_path(self, p):
        self._model_exposures_file_path = self.intermediate_files['model_exposures'] = p

    @property
    def keys_file_path(self):
        """
        Oasis keys file path property.

            :getter: Gets the file path
            :setter: Sets the current file path to the specified file path
        """
        return self._keys_file_path

    @keys_file_path.setter
    def keys_file_path(self, p):
        self._keys_file_path = self.intermediate_files['keys'] = p

    @property
    def keys_errors_file_path(self):
        """
        Oasis keys errors file path property.

            :getter: Gets the file path
            :setter: Sets the current file path to the specified file path
        """
        return self._keys_errors_file_path

    @keys_errors_file_path.setter
    def keys_errors_file_path(self, p):
        self._keys_errors_file_path = self.intermediate_files['keys_errors'] = p

    @property
    def items_file_path(self):
        """
        Oasis items file path property.

            :getter: Gets the file path
            :setter: Sets the current file path to the specified file path
        """
        return self._items_file_path

    @items_file_path.setter
    def items_file_path(self, p):
        self._items_file_path = self.gul_files['items'] = self.oasis_files['items'] = p

    @property
    def coverages_file_path(self):
        """
        Oasis coverages file path property.

            :getter: Gets the file path
            :setter: Sets the current file path to the specified file path
        """
        return self._coverages_file_path

    @coverages_file_path.setter
    def coverages_file_path(self, p):
        self._coverages_file_path = self.gul_files['coverages'] = self.oasis_files['coverages'] = p

    @property
    def gulsummaryxref_file_path(self):
        """
        GUL summary file path property.

            :getter: Gets the file path
            :setter: Sets the current file path to the specified file path
        """
        return self._gulsummaryxref_file_path

    @gulsummaryxref_file_path.setter
    def gulsummaryxref_file_path(self, p):
        self._gulsummaryxref_file_path = self.gul_files['gulsummaryxref'] = self.oasis_files['gulsummaryxref'] = p

    @property
    def fm_policytc_file_path(self):
        """
        FM policy T & C file property.

            :getter: Gets the actual file object
            :setter: Sets the file to the specified file object
        """
        return self._fm_policytc_file_path

    @fm_policytc_file_path.setter
    def fm_policytc_file_path(self, p):
        self._fm_policytc_file_path = self.fm_files['fm_policytc'] = self.oasis_files['fm_policytc'] = p

    @property
    def fm_profile_file_path(self):
        """
        FM profile file property.

            :getter: Gets the actual file object
            :setter: Sets the file to the specified file object
        """
        return self._fm_profile_file_path

    @fm_profile_file_path.setter
    def fm_profile_file_path(self, p):
        self._fm_profile_file_path = self.fm_files['fm_profile'] = self.oasis_files['fm_profile'] = p

    @property
    def fm_programme_file_path(self):
        """
        FM programme file property.

            :getter: Gets the actual file object
            :setter: Sets the file to the specified file object
        """
        return self._fm_programme_file_path

    @fm_programme_file_path.setter
    def fm_programme_file_path(self, p):
        self._fm_programme_file_path = self.fm_files['fm_programme'] = self.oasis_files['fm_programme'] = p

    @property
    def fm_xref_file_path(self):
        """
        FM xref file property.

            :getter: Gets the actual file object
            :setter: Sets the file to the specified file object
        """
        return self._fm_xref_file_path

    @fm_xref_file_path.setter
    def fm_xref_file_path(self, p):
        self._fm_xref_file_path = self.fm_files['fm_xref'] = self.oasis_files['fm_xref'] = p

    @property
    def fmsummaryxref_file_path(self):
        """
        FM fmsummaryxref file property.

            :getter: Gets the actual file object
            :setter: Sets the file to the specified file object
        """
        return self._fmsummaryxref_file_path

    @fmsummaryxref_file_path.setter
    def fmsummaryxref_file_path(self, p):
        self._fmsummaryxref_file_path = self.fm_files['fmsummaryxref'] = self.oasis_files['fmsummaryxref'] = p

    @property
    def source_files(self):
        """
        Oasis source files set property - getter only.

            :getter: Gets the complete set of paths of the Oasis source files,
            including source loc. and acc. files.
        """
        return self._source_files

    @property
    def intermediate_files(self):
        """
        Oasis intermediate files set property - getter only.

            :getter: Gets the complete set of paths of the Oasis intermediate
            files, including canonical and model exposures files, and keys
            and keys errors files.
        """
        return self._intermediate_files

    @property
    def gul_files(self):
        """
        Oasis GUL files set property - getter only.

            :getter: Gets the complete set of paths of the generated Oasis
            GUL files, including  ``items.csv``, ``coverages.csv``, `gulsummaryxref.csv`.
        """
        return self._gul_files

    @property
    def fm_files(self):
        """
        Oasis FM files set property - getter only.

            :getter: Gets the complete set of generated FM files, including
                     ``fm_policytc.csv``, ``fm_profile.csv``, `fm_programme.csv`,
                     ``fm_xref.csv``, ``fmsummaryxref.csv``.
        """
        return self._fm_files

    @property
    def oasis_files(self):
        """
        Oasis GUL + FM files set property - getter only.

            :getter: Gets the complete set of generated GUL + FM files,
                     including ``items.csv`, ``coverages.csv``,
                     ``gulsummaryxref.csv``, ``fm_policytc.csv``,
                     ``fm_profile.csv``, `fm_programme.csv`,
                     ``fm_xref.csv``, ``fmsummaryxref.csv``.
        """
        return self._oasis_files


    def clear(self, files_subsets=None):
        """
        Clears file path attributes in the pipeline.

            :param files_subsets: List of files subset labels (if empty)
                                  then all files subsets are cleared)::

                'source': source exposures file path + source accounts
                                file path
                'intermediate': canonical exposures file path, model
                                      exposures file path, keys file path,
                                      keys errors file path
                'gul': GUL files
                'fm': FM files
                `oasis`: GUL files + GM files

            :type files_subsets: list
        """
        if not files_subsets:
            filenames = chain(self.source_files, self.intermediate_files, self.oasis_files)
        else:
            filenames = chain(fn for fsb in files_subsets for fn in getattr(self, '{}_files'.format(fsb)))

        map(
            lambda p: setattr(self, '{}_file_path'.format(p), None),
            filenames
        )



# -*- coding: utf-8 -*-

__all__ = [
    'OasisExposuresManagerInterface',
    'OasisExposuresManager'
]

import copy
import io
import itertools
import json
import logging
import multiprocessing
import os
import shutil
import six
import sys
import time

import pandas as pd

from interface import (
    Interface,
    implements,
)


from ..keys.lookup import OasisLookupFactory
from ..utils.concurrency import (
    multiprocess,
    multithread,
    Task,
)
from ..utils.exceptions import OasisException
from ..utils.fm import (
    unified_canonical_fm_profile_by_level_and_term_group,
    get_fm_terms_by_level_as_list,
    get_policytc_ids,
)
from ..utils.metadata import OASIS_FM_LEVELS
from ..utils.values import get_utctimestamp
from ..models import OasisModel
from .pipeline import OasisFilesPipeline
from .csv_trans import Translator


class OasisExposuresManagerInterface(Interface):  # pragma: no cover
    """
    Interface class form managing a collection of exposures.

    :param oasis_models: A list of Oasis model objects with resources provided in the model objects'
        resources dictionaries.
    :type oasis_models: ``list(OasisModel)``
    """

    def __init__(self, oasis_models=None):
        """
        Class constructor.

        :param oasis_models: An optional list of Oasis model objects
        :type oasis_models: list
        """
        pass

    def add_model(self, oasis_model):
        """
        Adds Oasis model object to the manager and sets up its resources.

        :param oasis_model: An Oasis model object 
        :type oasis_model: oasislmf.models.model.OasisModel
        """
        pass

    def delete_model(self, oasis_model):
        """
        Deletes an existing Oasis model object in the manager.

        :param oasis_model: An Oasis model object 
        :type oasis_model: oasislmf.models.model.OasisModel
        """
        pass

    def transform_source_to_canonical(self, oasis_model=None, **kwargs):
        """
        Transforms a source exposures/locations for a given ``oasis_model``
        or set of keyword arguments to a canonical/standard Oasis format.

        All the required resources must be provided either in the model object
        resources dict or the keyword arguments.

        It is up to the specific implementation of this class of how these
        resources will be named and how they will be used to
        effect the transformation.

        The transform is generic by default, but could be supplier specific if
        required.

        :param oasis_model: An Oasis model object 
        :type oasis_model: oasislmf.models.model.OasisModel

        :param kwargs: Optional keyword arguments
        """
        pass

    def transform_canonical_to_model(self, oasis_model=None, **kwargs):
        """
        Transforms the canonical exposures/locations for a given ``oasis_model``
        or set of keyword arguments object to a format suitable for an Oasis
        model keys lookup service.

        All the required resources must be provided either in the model object
        resources dict or the keyword arguments.

        It is up to the specific implementation of this class of how these
        resources will be named and how they will be used to
        effect the transformation.

        :param oasis_model: An Oasis model object 
        :type oasis_model: oasislmf.models.model.OasisModel

        :param kwargs: Optional keyword arguments
        """
        pass

    def get_keys(self, oasis_model=None, **kwargs):
        """
        Generates the Oasis keys and keys error files for a given
        ``oasis_model`` or set of keyword arguments.

        The keys file is a CSV file containing keys lookup information for
        locations with successful lookups, and has the following headers::

            LocID,PerilID,CoverageTypeID,AreaPerilID,VulnerabilityID

        while the keys error file is a CSV file containing keys lookup
        information for locations with unsuccessful lookups (failures,
        no matches) and has the following headers::

            LocID,PerilID,CoverageTypeID,Message

        All the required resources must be provided either in the model object
        resources dict or the keyword arguments.

        It is up to the specific implementation of this class of how these
        resources will be named and how they will be used to
        effect the transformation.

        A "standard" implementation should use the lookup service factory
        class in ``oasis_utils`` (a submodule of `omdk`) namely

            ``oasis_utils.oasis_keys_lookup_service_utils.KeysLookupServiceFactory``

        :param oasis_model: An Oasis model object 
        :type oasis_model: oasislmf.models.model.OasisModel

        :param kwargs: Optional keyword arguments
        """
        pass

    def load_canonical_exposures_profile(self, oasis_model=None, **kwargs):
        """
        Loads a JSON string or JSON file representation of the canonical
        exposures profile for a given model.

        :param oasis_model: An Oasis model object 
        :type oasis_model: oasislmf.models.model.OasisModel

        :param kwargs: Optional keyword arguments
        """
        pass

    def load_canonical_accounts_profile(self, oasis_model=None, **kwargs):
        """
        Loads a JSON string or JSON file representation of the canonical
        accounts profile for a given model.

        :param oasis_model: An Oasis model object 
        :type oasis_model: oasislmf.models.model.OasisModel

        :param kwargs: Optional keyword arguments
        """
        pass

    def load_fm_aggregation_profile(self, oasis_model=None, **kwargs):
        """
        Loads a JSON string or JSON file representation of the FM
        aggregation profile for a given model.

        :param oasis_model: An Oasis model object 
        :type oasis_model: oasislmf.models.model.OasisModel

        :param kwargs: Optional keyword arguments
        """
        pass

    def generate_gul_items(self, canonical_exposures_profile, canonical_exposures_df, keys_df, **kwargs):
        """
        Generates GUL items.

        :param canonical_exposures_profile: Canonical exposures profile
        :type canonical_exposures_profile: dict

        :param canonical_exposures_df: Canonical exposures
        :type canonical_exposures_df: pandas.DataFrame

        :param keys_df: Keys
        :type keys_df: pandas.DataFrame
        """
        pass

    def generate_fm_items(self, canonical_exposures_df, gul_items_df, canonical_exposures_profile, canonical_accounts_profile, canonical_accounts_df, fm_agg_profile, **kwargs):
        """
        Generates FM items.

        :param canonical_exposures_df: Canonical exposures
        :type canonical_exposures_df: pandas.DataFrame

        :param gul_items_df: GUL items
        :type gul_items_df: pandas.DataFrame

        :param canonical_exposures_profile: Canonical exposures profile
        :type canonical_exposures_profile: dict

        :param canonical_accounts_profile: Canonical accounts profile
        :type canonical_accounts_profile: dict

        :param canonical_accounts_df: Canonical accounts
        :param canonical_accounts_df: pandas.DataFrame

        :param fm_agg_profile: FM aggregation profile
        :param fm_agg_profile: dict
        """
        pass

    def load_gul_items(self, canonical_exposures_profile, canonical_exposures_file_path, keys_file_path, **kwargs):
        """
        Loads GUL items generated by ``generate_gul_items`` into a static
        structure such as a pandas dataframe.

        :param canonical_exposures_profile: Canonical exposures profile
        :type canonical_exposures_profile: dict

        :param canonical_exposures_file_path: Canonical exposures file path
        :type canonical_exposures_file_path: str

        :param keys_file_path: Keys file path
        :type keys_file_path: str
        """
        pass

    def load_fm_items(self, canonical_exposures_df, gul_items_df, canonical_exposures_profile, canonical_accounts_profile, canonical_accounts_file_path, fm_agg_profile, **kwargs):
        """
        Loads FM items generated by ``generate_fm_items`` into a static
        structure such as a pandas dataframe.

        :param canonical_exposures_df: Canonical exposures
        :type canonical_exposures_df: pandas.DataFrame

        :param gul_items_df: GUL items
        :type gul_items_df: pandas.DataFrame

        :param canonical_exposures_profile: Canonical exposures profile
        :type canonical_exposures_profile: dict

        :param canonical_accounts_profile: Canonical accounts profile
        :type canonical_accounts_profile: dict

        :param canonical_accounts_file_path: Canonical accounts file path
        :param canonical_accounts_file_path: str

        :param fm_agg_profile: FM aggregation profile
        :param fm_agg_profile: dict
        """
        pass

    def write_gul_files(self, oasis_model=None, **kwargs):
        """
        Writes Oasis GUL files for a given ``oasis_model`` or set of keyword
        arguments.

        The required resources must be provided either via the model object
        resources dict or the keyword arguments.

        :param oasis_model: An Oasis model object 
        :type oasis_model: oasislmf.models.model.OasisModel

        :param kwargs: Optional keyword arguments
        """
        pass

    def write_fm_files(self, oasis_model=None, **kwargs):
        """
        Writes Oasis FM files for a given ``oasis_model`` or set of keyword
        arguments.

        The required resources must be provided either via the model object
        resources dict or the keyword arguments.

        :param oasis_model: An Oasis model object 
        :type oasis_model: oasislmf.models.model.OasisModel

        :param kwargs: Optional keyword arguments
        """
        pass

    def write_oasis_files(self, oasis_model=None, fm=False, **kwargs):
        """
        Writes the full set of Oasis files, which includes GUL files and
        possibly also the FM files (if ``fm`` is ``True``) for a given
        ``oasis_model`` or set of keyword arguments.

        The required resources must be provided either via the model object
        resources dict or the keyword arguments.

        :param oasis_model: An Oasis model object 
        :type oasis_model: oasislmf.models.model.OasisModel

        :param kwargs: Optional keyword arguments
        """
        pass

    def create_model(self, model_supplier_id, model_id, model_version, resources=None):
        """
        Creates an Oasis model object, with attached resources if a resources
        dict was provided.

        :param model_supplier_id: The model supplier ID
        :type model_supplier_id: str

        :param model_id: The model ID
        :type model_id: str

        :param model_version: The model version string
        :type model_version: str

        :param resources: Optional dictionary of model resources
        :type resources: dict
        """
        pass


class OasisExposuresManager(implements(OasisExposuresManagerInterface)):

    def __init__(self, oasis_models=None):
        self.logger = logging.getLogger()

        self.logger.debug('Exposures manager {} initialising'.format(self))

        self.logger.debug('Adding models')
        self._models = {}

        self.add_models(oasis_models)

        self.logger.debug('Exposures manager {} finished initialising'.format(self))

    def add_model(self, oasis_model):
        """
        Adds model to the manager and sets up its resources.
        """
        self._models[oasis_model.key] = oasis_model

        return oasis_model

    def add_models(self, oasis_models):
        """
        Adds a list of Oasis model objects to the manager.
        """
        for model in oasis_models or []:
            self.add_model(model)

    def delete_model(self, oasis_model):
        """
        Deletes an existing Oasis model object in the manager.
        """
        if oasis_model.key in self._models:
            oasis_model.resources['oasis_files_pipeline'].clear()

            del self._models[oasis_model.key]

    def delete_models(self, oasis_models):
        """
        Deletes a list of existing Oasis model objects in the manager.
        """
        for model in oasis_models:
            self.delete_model(model)

    @property
    def keys_lookup_factory(self):
        """
        Keys lookup service factory property - getter only.

            :getter: Gets the current keys lookup service factory instance
        """
        return self._keys_lookup_factory

    @property
    def models(self, key=None):
        """
        Model objects dictionary property.

            :getter: Gets the model in the models dict using the optional
                     ``key`` argument. If ``key`` is not given then the dict
                     is returned.

            :setter: Sets the value of the optional ``key`` in the models dict
                     to ``val`` where ``val`` is assumed to be an Oasis model
                     object (``omdk.OasisModel.OasisModel``).

                     If no ``key`` is given then ``val`` is assumed to be a new
                     models dict and is used to replace the existing dict.

            :deleter: Deletes the value of the optional ``key`` in the models
                      dict. If no ``key`` is given then the entire existing
                      dict is cleared.
        """
        return self._models[key] if key else self._models

    @models.setter
    def models(self, val, key=None):
        if key:
            self._models.update({key:val})
        else:
            self._models.update(val)

    @models.deleter
    def models(self):
        self._models.clear()

    def transform_source_to_canonical(self, oasis_model=None, source_type='exposures', **kwargs):
        """
        Transforms a canonical exposures/locations file for a given
        ``oasis_model`` object to a canonical/standard Oasis format.

        It can also transform a source accounts file to a canonical accounts
        file, if the optional argument ``source_type`` has the value of ``accounts``.
        The default ``source_type`` is ``exposures``.

        By default parameters supplied to this function fill be used if present
        otherwise they will be taken from the `oasis_model` resources dictionary
        if the model is supplied.

        :param oasis_model: An optional Oasis model object
        :type oasis_model: ``oasislmf.models.model.OasisModel``

        :param source_exposures_file_path: Source exposures file path (if ``source_type`` is ``exposures``)
        :type source_exposures_file_path: str

        :param source_exposures_validation_file_path: Source exposures validation file (if ``source_type`` is ``exposures``)
        :type source_exposures_validation_file_path: str

        :param source_to_canonical_exposures_transformation_file_path: Source exposures transformation file (if ``source_type`` is ``exposures``)
        :type source_to_canonical_exposures_transformation_file_path: str

        :param canonical_exposures_file_path: Path to the output canonical exposure file (if ``source_type`` is ``exposures``)
        :type canonical_exposures_file_path: str

        :param source_accounts_file_path: Source accounts file path (if ``source_type`` is ``accounts``)
        :type source_exposures_file_path: str

        :param source_accounts_validation_file_path: Source accounts validation file (if ``source_type`` is ``accounts``)
        :type source_exposures_validation_file_path: str

        :param source_to_canonical_accounts_transformation_file_path: Source accounts transformation file (if ``source_type`` is ``accounts``)
        :type source_to_canonical_accounts_transformation_file_path: str

        :param canonical_accounts_file_path: Path to the output canonical accounts file (if ``source_type`` is ``accounts``)
        :type canonical_accounts_file_path: str

        :return: The path to the output canonical file
        """
        kwargs = self._process_default_kwargs(oasis_model=oasis_model, **kwargs)

        input_file_path = os.path.abspath(kwargs['source_accounts_file_path']) if source_type == 'accounts' else os.path.abspath(kwargs['source_exposures_file_path'])
        transformation_file_path = os.path.abspath(kwargs['source_to_canonical_accounts_transformation_file_path']) if source_type == 'accounts' else os.path.abspath(kwargs['source_to_canonical_exposures_transformation_file_path'])

        try:
            validation_file_path = kwargs['source_accounts_validation_file_path'] if source_type == 'accounts' else kwargs.get['source_exposures_validation_file_path']
        except (KeyError, TypeError) as e:
            validation_file_path = None

        output_file_path = os.path.abspath(kwargs['canonical_accounts_file_path']) if source_type == 'accounts' else os.path.abspath(kwargs['canonical_exposures_file_path'])

        translator = Translator(input_file_path, output_file_path, transformation_file_path, xsd_path=validation_file_path, append_row_nums=True)

        translator()

        if oasis_model:
            if source_type == 'accounts':
                oasis_model.resources['oasis_files_pipeline'].canonical_accounts_file_path = output_file_path
            else:
                oasis_model.resources['oasis_files_pipeline'].canonical_exposures_file_path = output_file_path

        return output_file_path

    def transform_canonical_to_model(self, oasis_model=None, **kwargs):
        """
        Transforms the canonical exposures/locations file for a given
        ``oasis_model`` object to a format suitable for an Oasis model keys
        lookup service.

        By default parameters supplied to this function fill be used if present
        otherwise they will be taken from the `oasis_model` resources dictionary
        if the model is supplied.

        :param oasis_model: The model to get keys for
        :type oasis_model: ``oasislmf.models.model.OasisModel``

        :param canonical_exposures_file_path: Path to the canonical exposures file
        :type canonical_exposures_file_path: str

        :param canonical_exposures_validation_file_path: Path to the exposure validation file
        :type canonical_exposures_validation_file_path: str

        :param canonical_to_model_exposures_transformation_file_path: Path to the exposure transformation file
        :type canonical_to_model_exposures_transformation_file_path: str

        :param model_exposures_file_path: Path to the output model exposure file
        :type model_exposures_file_path: str

        :return: The path to the output model exposure file
        """
        kwargs = self._process_default_kwargs(oasis_model=oasis_model, **kwargs)

        input_file_path = os.path.abspath(kwargs.get('canonical_exposures_file_path'))
        transformation_file_path = os.path.abspath(kwargs.get('canonical_to_model_exposures_transformation_file_path'))

        try:
            validation_file_path = kwargs['canonical_exposures_validation_file_path']
        except (KeyError, TypeError) as e:
            validation_file_path = None

        output_file_path = os.path.abspath(kwargs.get('model_exposures_file_path'))

        translator = Translator(input_file_path, output_file_path, transformation_file_path, xsd_path=validation_file_path, append_row_nums=False)

        translator()

        if oasis_model:
            oasis_model.resources['oasis_files_pipeline'].model_exposures_file_path = output_file_path

        return output_file_path

    def load_canonical_exposures_profile(self, oasis_model=None, **kwargs):
        """
        Loads a JSON string or JSON file representation of the canonical
        exposures profile for a given ``oasis_model``, stores this in the
        model object's resources dict, and returns the object.
        """
        kwargs = self._process_default_kwargs(oasis_model=oasis_model, **kwargs)

        profile_json = kwargs.get('canonical_exposures_profile_json')
        profile_json_path = kwargs.get('canonical_exposures_profile_json_path')

        profile = None
        if profile_json:
            profile = json.loads(profile_json)
        elif profile_json_path:
            with io.open(profile_json_path, 'r', encoding='utf-8') as f:
                profile = json.load(f)

        if oasis_model:
            oasis_model.resources['canonical_exposures_profile'] = profile

        return profile

    def load_canonical_accounts_profile(self, oasis_model=None, **kwargs):
        """
        Loads a JSON string or JSON file representation of the canonical
        exposures profile for a given ``oasis_model``, stores this in the
        model object's resources dict, and returns the object.
        """
        kwargs = self._process_default_kwargs(oasis_model=oasis_model, **kwargs)

        profile_json = kwargs.get('canonical_accounts_profile_json')
        profile_json_path = kwargs.get('canonical_accounts_profile_json_path')

        profile = None
        if profile_json:
            profile = json.loads(profile_json)
        elif profile_json_path:
            with io.open(profile_json_path, 'r', encoding='utf-8') as f:
                profile = json.load(f)

        if oasis_model:
            oasis_model.resources['canonical_accounts_profile'] = profile

        return profile

    def load_fm_aggregation_profile(self, oasis_model=None, **kwargs):
        """
        Loads a JSON string or JSON file representation of the FM
        aggregation profile for a given ``oasis_model``, stores this in the
        model object's resources dict, and returns the object.
        """
        kwargs = self._process_default_kwargs(oasis_model=oasis_model, **kwargs)

        profile_path = kwargs.get('fm_agg_profile_path')
        profile_json = kwargs.get('fm_agg_profile_json')

        profile = None
        if profile_json:
            profile = {int(k):v for k, v in six.iteritems(json.loads(profile_json))}
        elif profile_path:
            with io.open(profile_path, 'r', encoding='utf-8') as f:
                profile = {int(k):v for k, v in six.iteritems(json.load(f))}

        if oasis_model:
            oasis_model.resources['fm_agg_profile'] = profile

        return profile

    def load_lookup_config(self, oasis_model=None, **kwargs):
        """
        Loads a lookup config JSON string or file.
        """
        kwargs = self._process_default_kwargs(oasis_model=oasis_model, **kwargs)

        lookup = kwargs.get('lookup')

        if not lookup:
            return

        try:
            config = lookup.config
        except AttributeError:
            pass
        else:
            if oasis_model:
                oasis_model.resources['lookup_config'] = config

            return config

        config_json = kwargs.get('lookup_config_json')
        config_fp = kwargs.get('lookup_config_fp')

        config = None
        if config_json:
            config = json.loads(config_json)
        elif config_fp:
            with io.open(config_fp, 'r', encoding='utf-8') as f:
                config = json.load(f)

        if oasis_model:
            oasis_model.resources['lookup_config'] = config

        return config

    def get_keys(self, oasis_model=None, **kwargs):
        """
        Generates the Oasis keys and keys error files for a given model object.
        The keys file is a CSV file containing keys lookup information for
        locations with successful lookups, and has the following headers::

            LocID,PerilID,CoverageTypeID,AreaPerilID,VulnerabilityID

        while the keys error file is a CSV file containing keys lookup
        information for locations with unsuccessful lookups (failures,
        no matches) and has the following headers::

            LocID,PerilID,CoverageTypeID,Message

        By default it is assumed that all the resources required for the
        transformation are present in the model object's resources dict,
        if the model is supplied. These can be overridden by providing the
        relevant optional parameters.

        If no model is supplied then the optional paramenters must be
        supplied.

        If the model is supplied the result keGy file path is stored in the
        models ``file_pipeline.keyfile_path`` property.

        :param oasis_model: The model to get keys for
        :type oasis_model: ``OasisModel``

        :return: The path to the generated keys file
        """
        kwargs = self._process_default_kwargs(oasis_model=oasis_model, **kwargs)

        model_exposures_file_path = kwargs.get('model_exposures_file_path')
        lookup = kwargs.get('lookup')
        keys_file_path = kwargs.get('keys_file_path')
        keys_errors_file_path = kwargs.get('keys_errors_file_path')

        for p in (model_exposures_file_path, keys_file_path, keys_errors_file_path,):
            p = os.path.abspath(p) if p and not os.path.isabs(p) else p

        keys_file_path, _, keys_errors_file_path, _ = OasisLookupFactory().save_results(
            lookup,
            keys_file_path,
            errors_fp=keys_errors_file_path,
            model_exposures_fp=model_exposures_file_path
        )

        if oasis_model:
            oasis_model.resources['oasis_files_pipeline'].keys_file_path = keys_file_path
            oasis_model.resources['oasis_files_pipeline'].keys_errors_file_path = keys_errors_file_path

        return keys_file_path, keys_errors_file_path

    def _process_default_kwargs(self, oasis_model=None, **kwargs):
        if oasis_model:
            omr = oasis_model.resources
            ofp = omr['oasis_files_pipeline']

            kwargs.setdefault('fm', omr.get('fm') or kwargs.get('fm'))

            kwargs.setdefault('logger', omr.get('logger') or logging.getLogger())

            kwargs.setdefault('oasis_files_path', omr.get('oasis_files_path'))

            kwargs.setdefault('source_exposures_file_path', omr.get('source_exposures_file_path') or ofp.source_exposures_file_path)
            kwargs.setdefault('source_accounts_file_path', omr.get('source_accounts_file_path') or ofp.source_accounts_file_path)

            kwargs.setdefault('source_exposures_validation_file_path', omr.get('source_exposures_validation_file_path'))
            kwargs.setdefault('source_accounts_validation_file_path', omr.get('source_accounts_validation_file_path'))

            kwargs.setdefault('source_to_canonical_exposures_transformation_file_path', omr.get('source_to_canonical_exposures_transformation_file_path'))
            kwargs.setdefault('source_to_canonical_accounts_transformation_file_path', omr.get('source_to_canonical_accounts_transformation_file_path'))

            kwargs.setdefault('canonical_exposures_profile', omr.get('canonical_exposures_profile'))
            kwargs.setdefault('canonical_accounts_profile', omr.get('canonical_accounts_profile'))

            kwargs.setdefault('canonical_exposures_profile', omr.get('canonical_exposures_profile'))
            kwargs.setdefault('canonical_accounts_profile', omr.get('canonical_accounts_profile'))

            kwargs.setdefault('canonical_exposures_profile_json', omr.get('canonical_exposures_profile_json'))
            kwargs.setdefault('canonical_accounts_profile_json', omr.get('canonical_accounts_profile_json'))

            kwargs.setdefault('fm_agg_profile', omr.get('fm_agg_profile'))
            kwargs.setdefault('fm_agg_profile_path', omr.get('fm_agg_profile_path'))
            kwargs.setdefault('fm_agg_profile_json', omr.get('fm_agg_profile_json'))

            kwargs.setdefault('canonical_exposures_profile_json_path', omr.get('canonical_exposures_profile_json_path'))
            kwargs.setdefault('canonical_accounts_profile_json_path', omr.get('canonical_accounts_profile_json_path'))

            kwargs.setdefault('canonical_exposures_file_path', ofp.canonical_exposures_file_path)
            kwargs.setdefault('canonical_accounts_file_path', ofp.canonical_accounts_file_path)

            kwargs.setdefault('canonical_exposures_validation_file_path', omr.get('canonical_exposures_validation_file_path'))
            kwargs.setdefault('canonical_to_model_exposures_transformation_file_path', omr.get('canonical_to_model_exposures_transformation_file_path'))

            kwargs.setdefault('lookup_config', omr.get('lookup_config'))
            kwargs.setdefault('lookup_config_json', omr.get('lookup_config_json'))
            kwargs.setdefault('lookup_config_fp', omr.get('lookup_config_fp'))
            kwargs.setdefault('lookup', omr.get('lookup'))

            kwargs.setdefault('model_exposures_file_path', ofp.model_exposures_file_path)

            kwargs.setdefault('keys_file_path', ofp.keys_file_path)
            kwargs.setdefault('keys_errors_file_path', ofp.keys_errors_file_path)

            kwargs.setdefault('canonical_exposures_df', omr.get('canonical_exposures_df'))
            kwargs.setdefault('gul_items_df', omr.get('gul_items_df'))

            kwargs.setdefault('items_file_path', ofp.items_file_path)
            kwargs.setdefault('coverages_file_path', ofp.coverages_file_path)
            kwargs.setdefault('gulsummaryxref_file_path', ofp.gulsummaryxref_file_path)

            kwargs.setdefault('fm_items_df', omr.get('fm_items_df'))
            kwargs.setdefault('canonical_accounts_df', omr.get('canonical_accounts_df'))

            kwargs.setdefault('fm_policytc_file_path', ofp.fm_policytc_file_path)
            kwargs.setdefault('fm_profile_file_path', ofp.fm_profile_file_path)
            kwargs.setdefault('fm_policytc_file_path', ofp.fm_programme_file_path)
            kwargs.setdefault('fm_xref_file_path', ofp.fm_xref_file_path)
            kwargs.setdefault('fmsummaryxref_file_path', ofp.fmsummaryxref_file_path)

        return kwargs

    def generate_gul_items(
        self,
        canonical_exposures_profile,
        canonical_exposures_df,
        keys_df
    ):
        """
        Generates GUL items.

        :param canonical_exposures_profile: Canonical exposures profile
        :type canonical_exposures_profile: dict

        :param canonical_exposures_df: Canonical exposures data frame
        :type canonical_exposures_df: pandas.DataFrame

        :param keys_df: Keys file data frame
        :type keys_df: pandas.DataFrame
        """
        canexp_df = canonical_exposures_df

        cep = canonical_exposures_profile

        ufcp = unified_canonical_fm_profile_by_level_and_term_group(profiles=(cep,))

        if not ufcp:
            raise OasisException(
                'Canonical loc. profile is possibly missing FM term information: '
                'FM term definitions for TIV, limit, deductible, attachment and/or share.'
            )

        fm_levels = tuple(ufcp.keys())

        try:
            for df in [canexp_df, keys_df]:
                if not df.columns.contains('index'):
                    df['index'] = pd.Series(data=range(len(df)))

            if not str(canexp_df['row_id'].dtype).startswith('int'):
                canexp_df['row_id'] = canexp_df['row_id'].astype(int)

            if not str(keys_df['locid'].dtype).startswith('int'):
                keys_df['locid'] = keys_df['locid'].astype(int)

            oed_acc_col_repl = [{'accnumber': 'accntnum'}]
            for repl in oed_acc_col_repl:
                    canexp_df.rename(columns=repl, inplace=True)

            merged_df = pd.merge(canexp_df, keys_df, left_on='row_id', right_on='locid').drop_duplicates()
            merged_df['index'] = pd.Series(data=range(len(merged_df)), dtype=object)

            cov_level_id = fm_levels[0]

            tiv_terms = tuple(t for t in [ufcp[cov_level_id][gid].get('tiv') for gid in ufcp[cov_level_id]] if t)

            if not tiv_terms:
                raise OasisException('No TIV elements found in the canonical exposures profile - please check the canonical exposures (loc) profile')

            fm_terms = {
                tiv_tgid: {
                    term_type: (
                        ufcp[cov_level_id][tiv_tgid][term_type]['ProfileElementName'].lower() if ufcp[cov_level_id][tiv_tgid].get(term_type) else None
                    ) for term_type in ('deductible', 'deductiblemin', 'deductiblemax', 'limit', 'share',)
                } for tiv_tgid in ufcp[cov_level_id]
            }

            item_id = 0
            zero_tiv_items = 0
            positive_tiv_elements = lambda it: [t for t in tiv_terms if it.get(t['ProfileElementName'].lower()) and it[t['ProfileElementName'].lower()] > 0 and t['CoverageTypeID'] == it['coveragetypeid']] or [0]
            
            for it, ptiv in itertools.chain((it, ptiv) for _, it in merged_df.iterrows() for it, ptiv in itertools.product([it],positive_tiv_elements(it))):
                if ptiv == 0:
                    zero_tiv_items += 1
                    continue

                item_id += 1
                tiv_elm = ptiv['ProfileElementName'].lower()
                tiv = it[tiv_elm]
                tiv_tgid = ptiv['FMTermGroupID']

                yield {
                    'item_id': item_id,
                    'canexp_id': it['row_id'] - 1,
                    'peril_id': it['perilid'],
                    'coverage_type_id': it['coveragetypeid'],
                    'coverage_id': item_id,
                    'tiv_elm': tiv_elm,
                    'tiv': tiv,
                    'tiv_tgid': tiv_tgid,
                    'ded_elm': fm_terms[tiv_tgid].get('deductible'),
                    'ded_min_elm': fm_terms[tiv_tgid].get('deductiblemin'),
                    'ded_max_elm': fm_terms[tiv_tgid].get('deductiblemax'),
                    'lim_elm': fm_terms[tiv_tgid].get('limit'),
                    'shr_elm': fm_terms[tiv_tgid].get('share'),
                    'areaperil_id': it['areaperilid'],
                    'vulnerability_id': it['vulnerabilityid'],
                    'group_id': it['row_id'],
                    'summary_id': 1,
                    'summaryset_id': 1
                }
        except (AttributeError, KeyError, IndexError, TypeError, ValueError) as e:
            raise OasisException(e)
        else:
            if zero_tiv_items == len(merged_df):
                raise OasisException('All canonical exposure items have zero TIVs - please check the canonical exposures (loc.) file')

    def generate_fm_items(
        self,
        canonical_exposures_df,
        gul_items_df,
        canonical_exposures_profile,
        canonical_accounts_profile,
        canonical_accounts_df,
        fm_agg_profile
    ):
        """
        Generates FM items.

        :param canonical_exposures_df: Canonical exposures
        :type canonical_exposures_df: pandas.DataFrame

        :param gul_items_df: GUL items
        :type gul_items_df: pandas.DataFrame

        :param canonical_exposures_profile: Canonical exposures profile
        :type canonical_exposures_profile: dict

        :param canonical_accounts_profile: Canonical accounts profile
        :type canonical_accounts_profile: dict

        :param canonical_accounts_df: Canonical accounts
        :param canonical_accounts_df: pandas.DataFrame

        :param fm_agg_profile: FM aggregation profile
        :param fm_agg_profile: dict
        """
        cep = canonical_exposures_profile
        cap = canonical_accounts_profile
        
        canexp_df = canonical_exposures_df

        canacc_df = canonical_accounts_df

        for df in [canexp_df, gul_items_df, canacc_df]:
            if not df.columns.contains('index'):
                df['index'] = pd.Series(data=range(len(df)))

        oed_acc_col_repl = [{'accnumber': 'accntnum'}, {'polnumber': 'policynum'}]
        for repl in oed_acc_col_repl:
                canacc_df.rename(columns=repl, inplace=True)

        cangul_df = pd.merge(canexp_df, gul_items_df, left_on='index', right_on='canexp_id')
        cangul_df['index'] = pd.Series(data=cangul_df.index)

        keys = (
            'item_id', 'gul_item_id', 'peril_id', 'coverage_type_id', 'coverage_id',
            'canexp_id', 'canacc_id', 'policy_num', 'level_id', 'layer_id',
            'agg_id', 'policytc_id', 'deductible', 'deductible_min',
            'deductible_max', 'attachment', 'limit', 'share', 'calcrule_id', 'tiv_elm',
            'tiv', 'tiv_tgid', 'ded_elm', 'ded_min_elm', 'ded_max_elm',
            'lim_elm', 'shr_elm',
        )

        try:
            ufcp = unified_canonical_fm_profile_by_level_and_term_group(profiles=(cep, cap,))

            if not ufcp:
                raise OasisException(
                    'Canonical loc. and/or acc. profiles are possibly missing FM term information: '
                    'FM term definitions for TIV, limit, deductible and/or share.'
                )

            fmap = fm_agg_profile

            if not fmap:
                raise OasisException(
                    'FM aggregation profile is empty - this is required to perform aggregation'
                )

            fm_levels = tuple(ufcp.keys())

            cov_level_id = fm_levels[0]

            coverage_level_preset_data = list(zip(
                tuple(cangul_df.item_id.values),          # 1 - FM item ID
                tuple(cangul_df.item_id.values),          # 2 - GUL item ID
                tuple(cangul_df.peril_id.values),         # 3 - peril ID
                tuple(cangul_df.coverage_type_id.values), # 4 - coverage type ID
                tuple(cangul_df.coverage_id.values),      # 5 - coverage ID
                tuple(cangul_df.canexp_id.values),        # 6 - can. exp. DF index
                (-1,)*len(cangul_df),                     # 7 - can. acc. DF index
                (-1,)*len(cangul_df),                     # 8 - can. acc. policy num.
                (cov_level_id,)*len(cangul_df),           # 9 - coverage level ID
                (1,)*len(cangul_df),                      # 10 - layer ID
                (-1,)*len(cangul_df),                     # 11 - agg. ID
                tuple(cangul_df.tiv_elm.values),          # 12 - TIV element
                tuple(cangul_df.tiv.values),              # 13 -TIV value
                tuple(cangul_df.tiv_tgid.values),         # 14 -TIV element profile term group ID
                tuple(cangul_df.ded_elm.values),          # 15 -deductible element
                tuple(cangul_df.ded_min_elm.values),      # 16 -deductible min. element
                tuple(cangul_df.ded_max_elm.values),      # 17 -deductible max. element
                tuple(cangul_df.lim_elm.values),          # 18 -limit element
                tuple(cangul_df.shr_elm.values)           # 19 -share element
            ))

            get_canacc_item = lambda i: canacc_df[(canacc_df['accntnum'] == cangul_df[cangul_df['canexp_id']==coverage_level_preset_data[i][5]].iloc[0]['accntnum'])].iloc[0]

            get_canacc_id = lambda i: int(get_canacc_item(i)['index'])

            coverage_level_preset_items = {
                i: {
                    k:v for k, v in zip(
                        keys,
                        [i + 1, gul_item_id, peril_id, coverage_type_id, coverage_id, canexp_id, get_canacc_id(i), policy_num, level_id, layer_id, agg_id, -1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 12, tiv_elm, tiv, tiv_tgid, ded_elm, ded_min_elm, ded_max_elm, lim_elm, shr_elm]
                    )
                } for i, (item_id, gul_item_id, peril_id, coverage_type_id, coverage_id, canexp_id, _, policy_num, level_id, layer_id, agg_id, tiv_elm, tiv, tiv_tgid, ded_elm, ded_min_elm, ded_max_elm, lim_elm, shr_elm) in enumerate(coverage_level_preset_data)
            }

            num_cov_items = len(coverage_level_preset_items)

            preset_items = {
                level_id: (coverage_level_preset_items if level_id == cov_level_id else copy.deepcopy(coverage_level_preset_items)) for level_id in fm_levels
            }

            for i, (level_id, item_id, it) in enumerate(itertools.chain((level_id, k, v) for level_id in fm_levels[1:] for k, v in preset_items[level_id].items())):
                it['level_id'] = level_id
                it['item_id'] = num_cov_items + i + 1
                it['ded_elm'] = it['ded_min_elm'] = it['ded_max_elm'] = it['lim_elm'] = it['shr_elm'] = None

            num_sub_layer_level_items = sum(len(preset_items[level_id]) for level_id in fm_levels[:-1])
            layer_level = max(fm_levels)
            layer_level_items = copy.deepcopy(preset_items[layer_level])
            layer_level_min_idx = min(layer_level_items)

            layer_id = lambda i: list(
                canacc_df[canacc_df['accntnum'] == canacc_df.iloc[i]['accntnum']]['policynum'].values
            ).index(canacc_df.iloc[i]['policynum']) + 1

            for i, (canexp_id, canacc_id) in enumerate(
                itertools.chain((canexp_id, canacc_id) for canexp_id in layer_level_items for canexp_id, canacc_id in itertools.product(
                    [canexp_id],
                    canacc_df[canacc_df['accntnum'] == canacc_df.iloc[layer_level_items[canexp_id]['canacc_id']]['accntnum']]['index'].values)
                )
            ):
                it = copy.deepcopy(layer_level_items[canexp_id])
                it['item_id'] = num_sub_layer_level_items + i + 1
                it['layer_id'] = layer_id(canacc_id)
                it['canacc_id'] = canacc_id
                preset_items[layer_level][layer_level_min_idx + i] = it

            for it in (it for c in itertools.chain(six.itervalues(preset_items[k]) for k in preset_items) for it in c):
                it['policy_num'] = canacc_df.iloc[it['canacc_id']]['policynum']
                lfmaggkey = fmap[it['level_id']]['FMAggKey']
                for v in six.itervalues(lfmaggkey):
                    src = v['src'].lower()
                    if src in ['canexp', 'canacc']:
                        f = v['field'].lower()
                        it[f] = canexp_df.iloc[it['canexp_id']][f] if src == 'canexp' else canacc_df.iloc[it['canacc_id']][f]

            concurrent_tasks = (
                Task(get_fm_terms_by_level_as_list, args=(ufcp[level_id], fmap[level_id], preset_items[level_id], canexp_df.copy(deep=True), canacc_df.copy(deep=True),), key=level_id)
                for level_id in fm_levels
            )
            num_ps = min(len(fm_levels), multiprocessing.cpu_count())
            for it in multiprocess(concurrent_tasks, pool_size=num_ps):
                yield it
        except (AttributeError, KeyError, IndexError, TypeError, ValueError) as e:
            raise OasisException(e)

    def load_gul_items(self, canonical_exposures_profile, canonical_exposures_file_path, keys_file_path):
        """
        Loads GUL items generated by ``generate_gul_items`` into a static
        structure such as a pandas dataframe.

        :param canonical_exposures_profile: Canonical exposures profile
        :type canonical_exposures_profile: dict

        :param canonical_exposures_file_path: Canonical exposures file path
        :type canonical_exposures_file_path: str

        :param keys_file_path: Keys file path
        :type keys_file_path: str
        """
        cep = canonical_exposures_profile

        try:
            with io.open(canonical_exposures_file_path, 'r', encoding='utf-8') as cf, io.open(keys_file_path, 'r', encoding='utf-8') as kf:
                canexp_df, keys_df = pd.read_csv(cf, float_precision='high'), pd.read_csv(kf, float_precision='high')

            if len(canexp_df) == 0:
                raise OasisException('No canonical exposure items found - please check the canonical exposures (loc) file')

            if len(keys_df) == 0:
                raise OasisException('No keys items found - please check the model exposures (loc) file')

            canexp_df = canexp_df.where(canexp_df.notnull(), None)
            canexp_df.columns = canexp_df.columns.str.lower()
            canexp_df['index'] = pd.Series(data=canexp_df.index, dtype=int)

            keys_df = keys_df.where(keys_df.notnull(), None)
            keys_df.columns = keys_df.columns.str.lower()
            keys_df['index'] = pd.Series(data=keys_df.index, dtype=int)

            gul_items_df = pd.DataFrame(data=[it for it in self.generate_gul_items(cep, canexp_df, keys_df)], dtype=object)
            gul_items_df['index'] = pd.Series(data=gul_items_df.index, dtype=int)

            for col in gul_items_df.columns:
                if col.endswith('id'):
                    gul_items_df[col] = gul_items_df[col].astype(int)
                elif col == 'tiv':
                    gul_items_df[col] = gul_items_df[col].astype(float)
        except (IOError, MemoryError, OasisException, OSError, TypeError, ValueError) as e:
            raise OasisException(e)
            
        return gul_items_df, canexp_df


    def load_fm_items(
        self,
        canonical_exposures_df,
        gul_items_df,
        canonical_exposures_profile,
        canonical_accounts_profile,
        canonical_accounts_file_path,
        fm_agg_profile,
        reduced=True
    ):
        """
        Loads FM items generated by ``generate_fm_items`` into a static
        structure such as a pandas dataframe.

        :param canonical_exposures_df: Canonical exposures
        :type canonical_exposures_df: pandas.DataFrame

        :param gul_items_df: GUL items
        :type gul_items_df: pandas.DataFrame

        :param canonical_exposures_profile: Canonical exposures profile
        :type canonical_exposures_profile: dict

        :param canonical_accounts_profile: Canonical accounts profile
        :type canonical_accounts_profile: dict

        :param canonical_accounts_file_path: Canonical accounts file path
        :param canonical_accounts_file_path: str

        :param fm_agg_profile: FM aggregation profile
        :param fm_agg_profile: dict

        :param reduced: Whether to generate only FM items with not all zero
                        values for limit, deductible and share. By default ``True``
        :param reduced: bool
        """
        canexp_df = canonical_exposures_df

        cep = canonical_exposures_profile
        cap = canonical_accounts_profile

        fmap = fm_agg_profile

        try:
            with io.open(canonical_accounts_file_path, 'r', encoding='utf-8') as f:
                canacc_df = pd.read_csv(f, float_precision='high')

            if len(canacc_df) == 0:
                raise OasisException('No canonical accounts items')
            
            canacc_df = canacc_df.where(canacc_df.notnull(), None)
            canacc_df.columns = canacc_df.columns.str.lower()
            canacc_df['index'] = pd.Series(data=canacc_df.index, dtype=int)

            fm_items = [it for it in self.generate_fm_items(canexp_df, gul_items_df, cep, cap, canacc_df, fmap)]
            fm_items.sort(key=lambda it: it['item_id'])

            fm_items_df = pd.DataFrame(data=fm_items, dtype=object)
            fm_items_df['index'] = pd.Series(data=fm_items_df.index, dtype=int)

            bookend_fm_levels = (fm_items_df['level_id'].min(), fm_items_df['level_id'].max(),)

            if reduced:
                fm_items_df = fm_items_df[(fm_items_df['level_id'].isin(bookend_fm_levels)) | (fm_items_df['limit'] != 0) | (fm_items_df['deductible'] != 0) | (fm_items_df['deductible_min'] != 0) | (fm_items_df['deductible_max'] != 0) | (fm_items_df['share'] != 0)]

                fm_items_df['index'] = range(len(fm_items_df))

                fm_items_df['item_id'] = range(1, len(fm_items_df) + 1)

                level_ids = [l for l in set(fm_items_df['level_id'])]

                level_id = lambda i: level_ids.index(fm_items_df.iloc[i]['level_id']) + 1

                fm_items_df['level_id'] = fm_items_df['index'].apply(level_id)

            layer_level_id = fm_items_df['level_id'].max()

            policytc_ids = get_policytc_ids(fm_items_df)
            get_policytc_id = lambda i: [k for k in six.iterkeys(policytc_ids) if policytc_ids[k] == {k:fm_items_df.iloc[i][k] for k in ('limit', 'deductible', 'attachment', 'deductible_min', 'deductible_max', 'share', 'calcrule_id',)}][0]
            fm_items_df['policytc_id'] = fm_items_df['index'].apply(lambda i: get_policytc_id(i))

            for col in fm_items_df.columns:
                if col.endswith('id'):
                    fm_items_df[col] = fm_items_df[col].astype(int)
                elif col in ('tiv', 'limit', 'deductible', 'deductible_min', 'deductible_max', 'share',):
                    fm_items_df[col] = fm_items_df[col].astype(float)
        except (IOError, MemoryError, OasisException, OSError, TypeError, ValueError) as e:
            raise OasisException(e)

        return fm_items_df, canacc_df


    def write_items_file(self, gul_items_df, items_file_path):
        """
        Writes an items file.
        """
        try:
            gul_items_df.to_csv(
                columns=['item_id', 'coverage_id', 'areaperil_id', 'vulnerability_id', 'group_id'],
                path_or_buf=items_file_path,
                encoding='utf-8',
                chunksize=1000,
                index=False
            )
        except (IOError, OSError) as e:
            raise OasisException(e)

        return items_file_path

    def write_coverages_file(self, gul_items_df, coverages_file_path):
        """
        Writes a coverages file.
        """
        try:
            gul_items_df.to_csv(
                columns=['coverage_id', 'tiv'],
                path_or_buf=coverages_file_path,
                encoding='utf-8',
                chunksize=1000,
                index=False
            )
        except (IOError, OSError) as e:
            raise OasisException(e)

        return coverages_file_path

    def write_gulsummaryxref_file(self, gul_items_df, gulsummaryxref_file_path):
        """
        Writes a gulsummaryxref file.
        """
        try:
            gul_items_df.to_csv(
                columns=['coverage_id', 'summary_id', 'summaryset_id'],
                path_or_buf=gulsummaryxref_file_path,
                encoding='utf-8',
                chunksize=1000,
                index=False
            )
        except (IOError, OSError) as e:
            raise OasisException(e)

        return gulsummaryxref_file_path

    def write_fm_policytc_file(self, fm_items_df, fm_policytc_file_path):
        """
        Writes an FM policy T & C file.
        """
        try:
            fm_policytc_df = pd.DataFrame(
                columns=['layer_id', 'level_id', 'agg_id', 'policytc_id'],
                data=[key[:4] for key, _ in fm_items_df.groupby(['layer_id', 'level_id', 'agg_id', 'policytc_id', 'limit', 'deductible', 'share'])],
                dtype=object
            )
            fm_policytc_df.to_csv(
                path_or_buf=fm_policytc_file_path,
                encoding='utf-8',
                chunksize=1000,
                index=False
            )
        except (IOError, OSError) as e:
            raise OasisException(e)

        return fm_policytc_file_path

    def write_fm_profile_file(self, fm_items_df, fm_profile_file_path):
        """
        Writes an FM profile file.
        """
        try:
            cols = ['policytc_id', 'calcrule_id', 'limit', 'deductible', 'deductible_min', 'deductible_max', 'attachment', 'share']

            fm_profile_df = fm_items_df[cols]

            fm_profile_df = pd.DataFrame(
                columns=cols,
                data=[key for key, _ in fm_profile_df.groupby(cols)]
            )

            col_repl = [
                {'deductible': 'deductible1'},
                {'deductible_min': 'deductible2'},
                {'deductible_max': 'deductible3'},
                {'attachment': 'attachment1'},
                {'limit': 'limit1'},
                {'share': 'share1'}
            ]
            for repl in col_repl:
                fm_profile_df.rename(columns=repl, inplace=True)

            n = len(fm_profile_df)

            fm_profile_df['index'] = range(n)

            fm_profile_df['share2'] = fm_profile_df['share3'] = [0]*n

            fm_profile_df.to_csv(
                columns=['policytc_id','calcrule_id','deductible1', 'deductible2', 'deductible3', 'attachment1', 'limit1', 'share1', 'share2', 'share3'],
                path_or_buf=fm_profile_file_path,
                encoding='utf-8',
                chunksize=1000,
                index=False
            )
        except (IOError, OSError) as e:
            raise OasisException(e)

        return fm_profile_file_path

    def write_fm_programme_file(self, fm_items_df, fm_programme_file_path):
        """
        Writes a FM programme file.
        """
        try:
            fm_aggtree = {
                key:set(group['agg_id']) for key, group in fm_items_df[['level_id', 'agg_id']].groupby(['level_id'])
            }
            levels = sorted(fm_aggtree.keys())
            fm_aggtree[0] = fm_aggtree[levels[0]]
            levels = sorted(fm_aggtree.keys())

            data = [
                (a, second, b) for first, second in zip(levels, levels[1:]) for a, b in (
                    zip(fm_aggtree[first], fm_aggtree[second]) if (len(fm_aggtree[first]) == len(fm_aggtree[second]) and len(fm_aggtree[first]) > 1) else itertools.product(fm_aggtree[first], [list(fm_aggtree[second])[0]])
                )
            ]

            fm_programme_df = pd.DataFrame(columns=['from_agg_id', 'level_id', 'to_agg_id'], data=data, dtype=int)

            fm_programme_df.to_csv(
                path_or_buf=fm_programme_file_path,
                encoding='utf-8',
                chunksize=1000,
                index=False
            )
        except (IOError, OSError) as e:
            raise OasisException(e)

        return fm_programme_file_path

    def write_fm_xref_file(self, fm_items_df, fm_xref_file_path):
        """
        Writes a FM xref file.
        """
        try:
            data = [
                (i + 1, agg_id, layer_id) for i, (agg_id, layer_id) in enumerate(itertools.product(set(fm_items_df['agg_id']), set(fm_items_df['layer_id'])))
            ]

            fm_xref_df = pd.DataFrame(columns=['output', 'agg_id', 'layer_id'], data=data, dtype=int)

            fm_xref_df.to_csv(
                path_or_buf=fm_xref_file_path,
                encoding='utf-8',
                chunksize=1000,
                index=False
            )
        except (IOError, OSError) as e:
            raise OasisException(e)

        return fm_xref_file_path

    def write_fmsummaryxref_file(self, fm_items_df, fmsummaryxref_file_path):
        """
        Writes an FM summaryxref file.
        """
        try:
            data = [
                (i + 1, 1, 1) for i, _ in enumerate(itertools.product(set(fm_items_df['agg_id']), set(fm_items_df['layer_id'])))
            ]

            fmsummaryxref_df = pd.DataFrame(columns=['output', 'summary_id', 'summaryset_id'], data=data, dtype=int)

            fmsummaryxref_df.to_csv(
                path_or_buf=fmsummaryxref_file_path,
                encoding='utf-8',
                chunksize=1000,
                index=False
            )
        except (IOError, OSError) as e:
            raise OasisException(e)

        return fmsummaryxref_file_path

    def write_gul_files(self, oasis_model=None, **kwargs):
        """
        Writes the standard Oasis GUL files, namely::

            items.csv
            coverages.csv
            gulsummaryxref.csv
        """
        kwargs = self._process_default_kwargs(oasis_model=oasis_model, **kwargs)

        if oasis_model:
            omr = oasis_model.resources
            ofp = omr['oasis_files_pipeline']

        canonical_exposures_profile = kwargs.get('canonical_exposures_profile')
        canonical_exposures_file_path = kwargs.get('canonical_exposures_file_path')
        keys_file_path = kwargs.get('keys_file_path')
        
        gul_items_df, canexp_df = self.load_gul_items(canonical_exposures_profile, canonical_exposures_file_path, keys_file_path)

        if oasis_model:
            omr['canonical_exposures_df'] = canexp_df
            omr['gul_items_df'] = gul_items_df

        gul_files = (
            ofp.gul_files if oasis_model
            else {
                 'items': kwargs.get('items_file_path'),
                 'coverages': kwargs.get('coverages_file_path'),
                 'gulsummaryxref': kwargs.get('gulsummaryxref_file_path')
            }
        )

        concurrent_tasks = (
            Task(getattr(self, 'write_{}_file'.format(f)), args=(gul_items_df.copy(deep=True), gul_files[f],), key=f)
            for f in gul_files
        )
        num_ps = min(len(gul_files), multiprocessing.cpu_count())
        for _, _ in multithread(concurrent_tasks, pool_size=num_ps):
            pass

        return gul_files

    def write_fm_files(self, oasis_model=None, **kwargs):
        """
        Generate Oasis FM files, namely::

            fm_policytc.csv
            fm_profile.csv
            fm_programm.ecsv
            fm_xref.csv
            fm_summaryxref.csv
        """
        kwargs = self._process_default_kwargs(oasis_model=oasis_model, **kwargs)

        if oasis_model:
            omr = oasis_model.resources
            ofp = omr['oasis_files_pipeline']

            canexp_df, gul_items_df = omr['canonical_exposures_df'], omr['gul_items_df']
            canonical_exposures_profile = omr['canonical_exposures_profile']
            canonical_accounts_profile = omr['canonical_accounts_profile']
            canonical_accounts_file_path = ofp.canonical_accounts_file_path
            fm_agg_profile = omr['fm_agg_profile']
        else:
            canexp_df, gul_items_df = kwargs.get('canonical_exposures_df'), kwargs.get('gul_items_df')
            canonical_exposures_profile = kwargs.get('canonical_exposures_profile')
            canonical_accounts_profile = kwargs.get('canonical_accounts_profile')
            canonical_accounts_file_path = kwargs.get('canonical_accounts_file_path')
            fm_agg_profile = kwargs.get('fm_agg_profile')

        fm_items_df, canacc_df = self.load_fm_items(canexp_df, gul_items_df, canonical_exposures_profile, canonical_accounts_profile, canonical_accounts_file_path, fm_agg_profile)

        if oasis_model:
            omr['canonical_accounts_df'] = canacc_df
            omr['fm_items_df'] = fm_items_df

        fm_files = (
            ofp.fm_files if oasis_model
            else  {
                'fm_policytc': kwargs.get('fm_policytc_file_path'),
                'fm_profile': kwargs.get('fm_profile_file_path'),
                'fm_programme': kwargs.get('fm_programme_file_path'),
                'fm_xref': kwargs.get('fm_xref_file_path'),
                'fmsummaryxref': kwargs.get('fmsummaryxref_file_path')
            }
        )

        concurrent_tasks = (
            Task(getattr(self, 'write_{}_file'.format(f)), args=(fm_items_df.copy(deep=True), fm_files[f],), key=f)
            for f in fm_files
        )
        num_ps = min(len(fm_files), multiprocessing.cpu_count())
        n = len(fm_files)
        for _, _ in multithread(concurrent_tasks, pool_size=num_ps):
            pass

        return fm_files

    def write_oasis_files(self, oasis_model=None, fm=False, **kwargs):
        """
        Writes the Oasis files - GUL + FM (if ``fm`` is ``True``).
        """
        gul_files = self.write_gul_files(oasis_model=oasis_model, **kwargs)

        if not fm:
            return gul_files

        fm_files = self.write_fm_files(oasis_model=oasis_model, **kwargs)

        oasis_files = {k:v for k, v in itertools.chain(gul_files.items(), fm_files.items())}

        return oasis_files

    def clear_oasis_files_pipeline(self, oasis_model, **kwargs):
        """
        Clears the files pipeline for the given Oasis model object.
        """
        oasis_model.resources['oasis_files_pipeline'].clear()

        return oasis_model

    def start_oasis_files_pipeline(
        self,
        oasis_model=None,
        **kwargs
    ):
        """
        Starts the files pipeline for the given Oasis model object,
        which is the generation of the Oasis items, coverages and GUL summary
        files, and possibly the FM files, from the source exposures file,
        source accounts file, canonical exposures profile, and associated
        validation files and transformation files for the source and
        intermediate files (canonical exposures, model exposures).

        :param oasis_model: The Oasis model object
        :type oasis_model: oasislmf.models.model.OasisModel

        :param kwargs: Keyword arguments
        :type kwargs: dict

        :return: A dictionary of Oasis files (GUL + FM (if FM option indicated))
        """
        kwargs = self._process_default_kwargs(oasis_model=oasis_model, **kwargs)

        if oasis_model:
            omr = oasis_model.resources
            ofp = omr.get('oasis_files_pipeline') or OasisFilesPipeline(model_key=oasis_model.key)

        logger = kwargs.get('logger') or logging.getLogger()

        logger.info('\nChecking Oasis files directory exists for model')
        oasis_files_path = kwargs.get('oasis_files_path')

        if not oasis_files_path:
            raise OasisException('No Oasis files directory provided')
        elif not os.path.exists(oasis_files_path):
            raise OasisException('Oasis files directory {} does not exist on the filesystem.'.format(oasis_files_path))
        
        logger.info('\nOasis files directory: {}'.format(oasis_files_path))

        logger.info('\nChecking for source exposures file')
        source_exposures_file_path = kwargs.get('source_exposures_file_path')
        
        if not source_exposures_file_path:
            raise OasisException('No source exposures file path provided')
        elif not os.path.exists(source_exposures_file_path):
            raise OasisException("Source exposures file path {} does not exist on the filesysem".format(source_exposures_file_path))

        logger.info('\nFound source exposures file {source_exposures_file_path} - copying to Oasis files directory'.format(**kwargs))
        shutil.copy2(source_exposures_file_path, oasis_files_path)

        logger.info('\nLoading canonical exposures profile')
        canonical_exposures_profile = kwargs.get('canonical_exposures_profile') or self.load_canonical_exposures_profile(oasis_model=oasis_model, **kwargs)

        if canonical_exposures_profile is None:
            raise OasisException('No canonical exposures profile provided')

        logger.info('\nFound canonical exposures profile: {}'.format(canonical_exposures_profile))

        fm = kwargs.get('fm')

        source_accounts_file_path = None
        canonical_accounts_profile = None
        fm_agg_profile = None

        if fm:
            logger.info('\nChecking for source accounts file')
            source_accounts_file_path = kwargs.get('source_accounts_file_path')
            
            if not source_accounts_file_path:
                raise OasisException('FM option indicated but no source accounts file path provided in arguments or model resources')
            elif not os.path.exists(source_accounts_file_path):
                raise OasisException("Source accounts file path {} does not exist on the filesysem.".format(source_accounts_file_path))

            logger.info('\nFound source accounts file {source_accounts_file_path} - copying to Oasis files directory'.format(**kwargs))
            shutil.copy2(source_accounts_file_path, oasis_files_path)

            logger.info('\nLoading canonical accounts profile')
            canonical_accounts_profile = kwargs.get('canonical_accounts_profile') or self.load_canonical_accounts_profile(oasis_model=oasis_model, **kwargs)

            if canonical_accounts_profile is None:
                raise OasisException('FM option indicated but no canonical accounts profile provided')

            logger.info('\nLoaded canonical accounts profile: {}'.format(canonical_accounts_profile))

            logger.info('\nLoading FM aggregation profile')
            fm_agg_profile = kwargs.get('fm_agg_profile') or self.load_fm_aggregation_profile(oasis_model=oasis_model, **kwargs)

            if fm_agg_profile is None:
                raise OasisException('FM option indicated but no FM aggregation profile provided')

            logger.info('\nLoaded FM aggregation profile: {}'.format(fm_agg_profile))

        utcnow = get_utctimestamp(fmt='%Y%m%d%H%M%S')

        canonical_exposures_file_path = os.path.join(oasis_files_path, 'canexp-{}.csv'.format(utcnow))
        canonical_accounts_file_path = os.path.join(oasis_files_path, 'canacc-{}.csv'.format(utcnow))

        model_exposures_file_path = os.path.join(oasis_files_path, 'modexp-{}.csv'.format(utcnow))

        keys_file_path = os.path.join(oasis_files_path, 'oasiskeys-{}.csv'.format(utcnow))
        keys_errors_file_path = os.path.join(oasis_files_path, 'oasiskeys-errors-{}.csv'.format(utcnow))

        items_file_path = os.path.join(oasis_files_path, 'items.csv')
        coverages_file_path = os.path.join(oasis_files_path, 'coverages.csv')
        gulsummaryxref_file_path = os.path.join(oasis_files_path, 'gulsummaryxref.csv')

        fm_policytc_file_path = os.path.join(oasis_files_path, 'fm_policytc.csv')
        fm_profile_file_path = os.path.join(oasis_files_path, 'fm_profile.csv')
        fm_programme_file_path=os.path.join(oasis_files_path, 'fm_programme.csv')
        fm_xref_file_path = os.path.join(oasis_files_path, 'fm_xref.csv')
        fmsummaryxref_file_path = os.path.join(oasis_files_path, 'fmsummaryxref.csv')

        if oasis_model:
            ofp.source_exposures_file_path = source_exposures_file_path
            ofp.source_accounts_file_path = source_accounts_file_path

            ofp.canonical_exposures_file_path = canonical_exposures_file_path
            ofp.canonical_accounts_file_path = canonical_accounts_file_path

            ofp.model_exposures_file_path = model_exposures_file_path

            ofp.keys_file_path = keys_file_path
            ofp.keys_errors_file_path = keys_errors_file_path

            ofp.items_file_path = items_file_path
            ofp.coverages_file_path = coverages_file_path
            ofp.gulsummaryxref_file_path = gulsummaryxref_file_path

            ofp.fm_policytc_file_path = fm_policytc_file_path
            ofp.fm_profile_file_path = fm_profile_file_path
            ofp.fm_programme_file_path = fm_programme_file_path
            ofp.fm_xref_file_path = fm_xref_file_path
            ofp.fmsummaryxref_file_path = fmsummaryxref_file_path

        kwargs = self._process_default_kwargs(
            oasis_model=oasis_model,
            fm=fm,
            source_exposures_file_path=source_exposures_file_path,
            source_accounts_file_path=source_accounts_file_path,
            canonical_exposures_file_path=canonical_exposures_file_path,
            canonical_accounts_file_path=canonical_accounts_file_path,
            model_exposures_file_path=model_exposures_file_path,
            keys_file_path=keys_file_path,
            keys_errors_file_path=keys_errors_file_path,
            items_file_path=items_file_path,
            coverages_file_path=coverages_file_path,
            gulsummaryxref_file_path=gulsummaryxref_file_path,
            fm_policytc_file_path=fm_policytc_file_path,
            fm_profile_file_path=fm_profile_file_path,
            fm_programme_file_path=fm_programme_file_path,
            fm_xref_file_path=fm_xref_file_path,
            fmsummaryxref_file_path=fmsummaryxref_file_path
        )

        logger.info('\nWriting canonical exposures file {canonical_exposures_file_path}'.format(**kwargs))
        self.transform_source_to_canonical(oasis_model=oasis_model, **kwargs)

        if fm:
            logger.info('\nWriting canonical accounts file {canonical_accounts_file_path}'.format(**kwargs))
            self.transform_source_to_canonical(oasis_model=oasis_model, source_type='accounts', **kwargs)

        logger.info('\nWriting model exposures file {model_exposures_file_path}'.format(**kwargs))
        self.transform_canonical_to_model(oasis_model=oasis_model, **kwargs)

        logger.info('\nWriting keys file {keys_file_path} and keys errors file {keys_errors_file_path}'.format(**kwargs))

        self.get_keys(oasis_model=oasis_model, **kwargs)

        logger.info('\nWriting GUL files')
        gul_files = self.write_gul_files(oasis_model=oasis_model, **kwargs)

        if not fm:
            return gul_files

        logger.info('\nWriting FM files')
        fm_files = self.write_fm_files(oasis_model=oasis_model, **kwargs)

        oasis_files = ofp.oasis_files if oasis_model else {k:v for k, v in itertools.chain(gul_files.items(), fm_files.items())}

        return oasis_files

    def create_model(self, model_supplier_id, model_id, model_version, resources=None):
        model = OasisModel(
            model_supplier_id,
            model_id,
            model_version,
            resources=resources
        )

        # set default resources
        omr = model.resources

        omr.setdefault('oasis_files_path', os.path.abspath(os.path.join('Files', model.key.replace('/', '-'))))
        if not os.path.isabs(omr['oasis_files_path']):
            omr['oasis_files_path'] = os.path.abspath(omr['oasis_files_path'])

        ofp = OasisFilesPipeline(model_key=model.key)
        omr['oasis_files_pipeline'] = ofp

        if omr.get('canonical_exposures_profile') is None:
            self.load_canonical_exposures_profile(oasis_model=model)

        if omr.get('canonical_accounts_profile') is None:
            self.load_canonical_accounts_profile(oasis_model=model)

        if omr.get('fm_agg_profile') is None:
            self.load_fm_aggregation_profile(oasis_model=model)

        if omr.get('lookup') and omr.get('lookup_config') is None:
            self.load_lookup_config(oasis_model=model)

        self.add_model(model)

        return model



from __future__ import print_function, division

import glob
import os
import io
import re
import shutil
import sys
import tarfile
from contextlib import contextmanager
from distutils.log import INFO, WARN, ERROR
from distutils.spawn import find_executable
from tempfile import mkdtemp
from time import sleep

from setuptools import find_packages, setup, Command
from setuptools.command.install import install
from setuptools.command.develop import develop

try:
    from urllib.request import urlopen
    from urllib.error import URLError
except ImportError:
    from urllib2 import urlopen, URLError

KTOOLS_VERSION = '3.0.1'

SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))


def get_readme():
    with io.open(os.path.join(SCRIPT_DIR, 'README.md'), encoding='utf-8') as readme:
        return readme.read()


def get_install_requirements():
    with io.open(os.path.join(SCRIPT_DIR, 'requirements-package.in'), encoding='utf-8') as reqs:
        return reqs.readlines()


def get_version():
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    with io.open(os.path.join(SCRIPT_DIR, 'oasislmf', '__init__.py'), encoding='utf-8') as init_py:
        return re.search('__version__ = [\'"]([^\'"]+)[\'"]', init_py.read()).group(1)


@contextmanager
def temp_dir():
    d = mkdtemp()
    yield d
    shutil.rmtree(d, ignore_errors=True)


version = get_version()
reqs = get_install_requirements()
readme = get_readme()


class InstallKtoolsMixin(object):
    def fetch_ktools_tar(self, location, attempts=3, timeout=5, cooldown=1):
        self.announce('Retrieving ktools {}'.format(KTOOLS_VERSION), INFO)

        last_error = None
        request = None
        for i in range(attempts):
            try:
                request = urlopen('https://github.com/OasisLMF/ktools/archive/v{}.tar.gz'.format(KTOOLS_VERSION), timeout=timeout * 1000)
                break
            except URLError as e:
                self.announce('Failed to get ktools tar (attempt {})'.format(i + 1), WARN)
                last_error = e
                sleep(cooldown)
        else:
            self.announce('Failed to get ktools tar after {} attempts'.format(attempts), ERROR)
            if last_error:
                raise last_error

        with open(location, 'wb') as f:
            f.write(request.read())

    def unpack_tar(self, tar_location, extract_location):
        self.announce('Unpacking ktools', INFO)
        with tarfile.open(tar_location) as tar:
            if not os.path.exists(extract_location):
                os.makedirs(extract_location)
            tar.extractall(extract_location)

    def ktools_inpath(self):
        ktools_bin_subset = [
            'eve',
            'getmodel',
            'gulcalc',
            'fmcalc',
            'summarycalc',
        ]
        for ktools_bin in ktools_bin_subset:
            if find_executable(ktools_bin) is None:
                return False
        return True

    def build_ktools(self, extract_location):
        self.announce('Building ktools', INFO)
        build_dir = os.path.join(extract_location, 'ktools-{}'.format(KTOOLS_VERSION))

        exit_code = os.system('cd {build_dir} && ./autogen.sh && ./configure && make && make check'.format(build_dir=build_dir))
        if(exit_code is not 0):
            self.announce('Ktools build failed.\n', WARN)
            if (not self.ktools_inpath()):
                self.announce('Exisiting Ktools install not found.\n', WARN)
        return build_dir

    def add_ktools_to_path(self, build_dir):
        print('Installing ktools')

        if not os.path.exists(self.get_bin_dir()):
            os.makedirs(self.get_bin_dir())

        for p in glob.glob(os.path.join(build_dir, 'src', '*', '*')):
            split = p.split(os.path.sep)

            # if the file name is the same as the directory we have found a
            # component executable
            if split[-1] == split[-2]:
                component_path = os.path.join(self.get_bin_dir(), split[-1])
                shutil.copy(p, component_path)
                yield component_path

    def install_ktools(self):
        with temp_dir() as d:
            local_tar_path = os.path.join(d, 'ktools.tar.gz')
            local_extract_path = os.path.join(d, 'extracted')

            self.fetch_ktools_tar(local_tar_path)
            self.unpack_tar(local_tar_path, local_extract_path)
            build_dir = self.build_ktools(local_extract_path)
            self.ktools_components = list(self.add_ktools_to_path(build_dir))


class PostInstallKtools(InstallKtoolsMixin, install):
    command_name = 'install'
    user_options = install.user_options + [
        ('ktools', None, 'Only install ktools components'),
    ]
    boolean_options = install.boolean_options + ['ktools']

    def __init__(self, *args, **kwargs):
        self.ktools_components = []
        install.__init__(self, *args, **kwargs)

    def run(self):
        self.install_ktools()
        install.run(self)

    def get_outputs(self):
        return install.get_outputs(self) + self.ktools_components

    def get_bin_dir(self):
        return self.install_scripts


class PostDevelopKtools(InstallKtoolsMixin, develop):
    command_name = 'develop'
    user_options = develop.user_options
    boolean_options = develop.boolean_options

    def __init__(self, *args, **kwargs):
        self.ktools_components = []
        develop.__init__(self, *args, **kwargs)

    def run(self):
        self.install_ktools()
        develop.run(self)

    def get_outputs(self):
        return develop.get_outputs(self) + self.ktools_components

    def get_bin_dir(self):
        return self.script_dir


try:
    from wheel.bdist_wheel import bdist_wheel

    class BdistWheel(bdist_wheel):
        command_name = 'bdist_wheel'
        user_options = bdist_wheel.user_options

        def finalize_options(self):
            bdist_wheel.finalize_options(self)
            self.root_is_pure = False

        def get_tag(self):
            python, abi, plat = bdist_wheel.get_tag(self)
            python, abi = 'py2.py3', 'none'
            return python, abi, plat

except ImportError:
    BdistWheel = None

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


class Publish(Command):
    command_name = 'publish'
    user_options = [
        ('wheel', None, 'Publish the wheel'),
        ('sdist', None, 'Publish the sdist tar'),
        ('no-clean', None, 'Don\'t clean the build artifacts'),
        ('sign', None, 'Sign the artifacts using GPG')
    ]
    boolean_options = ['wheel', 'sdist']

    def initialize_options(self):
        self.wheel = False
        self.sdist = False
        self.no_clean = False
        self.sign = False

    def finalize_options(self):
        if not (self.wheel or self.sdist):
            self.announce('Either --wheel and/or --sdist must be provided', ERROR)
            sys.exit(1)

    def run(self):
        if os.system('pip freeze | grep twine'):
            self.announce('twine not installed.\nUse `pip install twine`.\nExiting.', WARN)
            sys.exit(1)

        if self.sdist:
            os.system('python setup.py sdist')

        if self.wheel:
            os.system('python setup.py bdist_wheel')

        if self.sign:
            for p in glob.glob('dist/*'):
                os.system('gpg --detach-sign -a {}'.format(p))

        os.system('twine upload dist/*')
        print('You probably want to also tag the version now:')
        print('  git tag -a {v} -m \'version {v}\''.format(v=version))
        print('  git push --tags')

        if not self.no_clean:
            shutil.rmtree('dist')
            shutil.rmtree('build')
            shutil.rmtree('oasislmf.egg-info')


setup(
    name='oasislmf',
    version=version,
    packages=find_packages(exclude=('tests', 'tests.*', 'tests.*.*')),
    include_package_data=True,
    package_data={
        '': [
            'requirements-package.in',
            'LICENSE',
        ],
        'oasislmf/_data/': ['*']
    },
    exclude_package_data={
        '': ['__pycache__', '*.py[co]'],
    },
    scripts=['bin/oasislmf'],
    license='BSD 3-Clause',
    description='Core loss modelling framework.',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/OasisLMF/oasislmf',
    author='Dan Bate (Wildfish), S Murthy (Oasis LMF)',
    author_email="Dan Bate <dan.bate@wildfish.com>,S Murthy <sandeep.murthy@oasislmf.org>",
    keywords='oasis lmf loss modeling framework',
    install_requires=reqs,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    cmdclass={
        'install': PostInstallKtools,
        'develop': PostDevelopKtools,
        'bdist_wheel': BdistWheel,
        'publish': Publish,
    },
)



from __future__ import unicode_literals

from collections import Counter

from oasislmf.exposures.oed import ALLOCATE_TO_ITEMS_BY_PREVIOUS_LEVEL_ALLOC_ID

import os
import io

RUNTYPE_GROUNDUP_LOSS = 'gul'
RUNTYPE_INSURED_LOSS = 'il'
RUNTYPE_REINSURANCE_LOSS = 'ri'

WAIT_PROCESSING_SWITCHES = {
    'full_uncertainty_aep': '-F',
    'wheatsheaf_aep': '-W',
    'sample_mean_aep': '-S',
    'full_uncertainty_oep': '-f',
    'wheatsheaf_oep': '-w',
    'sample_mean_oep': '-s',
    'wheatsheaf_mean_aep': '-M',
    'wheatsheaf_mean_oep': '-m',
}

def print_command(command_file, cmd):
    """
    Writes the supplied command to the end of the generated script

    :param cmd: The command to append
    """
    with io.open(command_file, "a", encoding='utf-8') as myfile:
        myfile.writelines(cmd + "\n")


def leccalc_enabled(lec_options):
    """
    Checks if leccalc is enabled in the leccalc options

    :param lec_options: The leccalc options from the analysis settings
    :type lec_options: dict

    :return: True is leccalc is enables, False otherwise.
    """
    for option in lec_options["outputs"]:
        if lec_options["outputs"][option]:
            return True
    return False


def do_post_wait_processing(runtype, analysis_settings, filename, process_counter):
    if '{}_summaries'.format(runtype) not in analysis_settings:
        return

    for summary in analysis_settings['{}_summaries'.format(runtype)]:
        if "id" in summary:
            summary_set = summary['id']
            if summary.get('aalcalc'):
                cmd = 'aalcalc -K{}_S{}_summaryaalcalc'.format(
                    runtype,
                    summary_set
                )

                process_counter['lpid_monitor_count'] += 1
                cmd = '{} > output/{}_S{}_aalcalc.csv'.format(cmd, runtype, summary_set)
                cmd = '{} & lpid{}=$!'.format(cmd, process_counter['lpid_monitor_count'])
                print_command(filename, cmd)

            if summary.get('lec_output'):
                leccalc = summary.get('leccalc', {})
                if leccalc and leccalc_enabled(leccalc):
                    cmd = 'leccalc {} -K{}_S{}_summaryleccalc'.format(
                        '-r' if leccalc.get('return_period_file') else '',
                        runtype,
                        summary_set
                    )

                    process_counter['lpid_monitor_count'] += 1
                    for option, active in sorted(leccalc['outputs'].items()):
                        if active:
                            switch = WAIT_PROCESSING_SWITCHES.get(option, '')
                            cmd = '{} {} output/{}_S{}_leccalc_{}.csv'.format(cmd, switch, runtype, summary_set,
                                                                              option)

                    cmd = '{} & lpid{}=$!'.format(cmd, process_counter['lpid_monitor_count'])
                    print_command(filename, cmd)


def do_fifos(action, runtype, analysis_settings, process_id, filename):
    summaries = analysis_settings.get('{}_summaries'.format(runtype))
    if not summaries:
        return

    print_command(filename, '{} fifo/{}_P{}'.format(action, runtype, process_id))
    print_command(filename, '')
    for summary in summaries:
        if 'id' in summary:
            summary_set = summary['id']
            print_command(filename, '{} fifo/{}_S{}_summary_P{}'.format(action, runtype, summary_set, process_id))

            if summary.get('eltcalc'):
                print_command(
                    filename,
                    '{} fifo/{}_S{}_summaryeltcalc_P{}'.format(action, runtype, summary_set, process_id)
                )
                print_command(
                    filename,
                    '{} fifo/{}_S{}_eltcalc_P{}'.format(action, runtype, summary_set, process_id)
                )

            if summary.get('summarycalc'):
                print_command(
                    filename,
                    '{} fifo/{}_S{}_summarysummarycalc_P{}'.format(action, runtype, summary_set, process_id)
                )
                print_command(
                    filename,
                    '{} fifo/{}_S{}_summarycalc_P{}'.format(action, runtype, summary_set, process_id)
                )

            if summary.get('pltcalc'):
                print_command(
                    filename,
                    '{} fifo/{}_S{}_summarypltcalc_P{}'.format(action, runtype, summary_set, process_id)
                )
                print_command(
                    filename,
                    '{} fifo/{}_S{}_pltcalc_P{}'.format(action, runtype, summary_set, process_id)
                )

    print_command(filename, '')


def create_workfolders(runtype, analysis_settings, filename):
    summaries = analysis_settings.get('{}_summaries'.format(runtype))
    if not summaries:
        return

    for summary in summaries:
        if 'id' in summary:
            summary_set = summary['id']
            if summary.get('lec_output'):
                if leccalc_enabled(summary['leccalc']):
                    print_command(filename, "mkdir work/{}_S{}_summaryleccalc".format(runtype, summary_set))

            if summary.get('aalcalc'):
                print_command(filename, 'mkdir work/{}_S{}_summaryaalcalc'.format(runtype, summary_set))


def remove_workfolders(runtype, analysis_settings, filename):
    print_command(filename, 'rm -rf work/kat')

    summaries = analysis_settings.get('{}_summaries'.format(runtype))
    if not summaries:
        return

    for summary in summaries:
        if 'id' in summary:
            summary_set = summary['id']
            if summary.get('lec_output'):
                if leccalc_enabled(summary['leccalc']):
                    print_command(filename, 'rm work/{}_S{}_summaryleccalc/*'.format(runtype, summary_set))
                    print_command(filename, 'rmdir work/{}_S{}_summaryleccalc'.format(runtype, summary_set))

            if summary.get('aalcalc'):
                print_command(filename, 'rm -rf work/{}_S{}_summaryaalcalc/*'.format(runtype, summary_set))
                print_command(filename, 'rmdir work/{}_S{}_summaryaalcalc'.format(runtype, summary_set))


def do_make_fifos(runtype, analysis_settings, process_id, filename):
    do_fifos('mkfifo', runtype, analysis_settings, process_id, filename)


def do_remove_fifos(runtype, analysis_settings, process_id, filename):
    do_fifos('rm', runtype, analysis_settings, process_id, filename)


def do_kats(runtype, analysis_settings, max_process_id, filename, process_counter):
    summaries = analysis_settings.get('{}_summaries'.format(runtype))
    if not summaries:
        return False

    anykats = False
    for summary in summaries:
        if 'id' in summary:
            summary_set = summary['id']

            if summary.get('eltcalc'):
                anykats = True

                cmd = 'kat'
                for process_id in range(1, max_process_id + 1):
                    cmd = '{} work/kat/{}_S{}_eltcalc_P{}'.format(cmd, runtype, summary_set, process_id)

                process_counter['kpid_monitor_count'] += 1
                cmd = '{} > output/{}_S{}_eltcalc.csv & kpid{}=$!'.format(cmd, runtype, summary_set,
                                                                          process_counter['kpid_monitor_count'])
                print_command(filename, cmd)

            if summary.get('pltcalc'):
                anykats = True

                cmd = 'kat'
                for process_id in range(1, max_process_id + 1):
                    cmd = '{} work/kat/{}_S{}_pltcalc_P{}'.format(cmd, runtype, summary_set, process_id)

                process_counter['kpid_monitor_count'] += 1
                cmd = '{} > output/{}_S{}_pltcalc.csv & kpid{}=$!'.format(cmd, runtype, summary_set,
                                                                          process_counter['kpid_monitor_count'])
                print_command(filename, cmd)

            if summary.get("summarycalc"):
                anykats = True

                cmd = 'kat'
                for process_id in range(1, max_process_id + 1):
                    cmd = '{} work/kat/{}_S{}_summarycalc_P{}'.format(cmd, runtype, summary_set, process_id)

                process_counter['kpid_monitor_count'] += 1
                cmd = '{} > output/{}_S{}_summarycalc.csv & kpid{}=$!'.format(cmd, runtype, summary_set,
                                                                              process_counter['kpid_monitor_count'])
                print_command(filename, cmd)

    return anykats


def do_summarycalcs(
    runtype, analysis_settings, process_id, filename, num_reinsurance_iterations=0):
    
    summaries = analysis_settings.get('{}_summaries'.format(runtype))
    if not summaries:
        return

    summarycalc_switch = '-f'
    if runtype == RUNTYPE_GROUNDUP_LOSS:
        summarycalc_switch = '-g'

    summarycalc_directory_switch = ""
    if runtype == RUNTYPE_REINSURANCE_LOSS:
        i = num_reinsurance_iterations
        summarycalc_directory_switch = "-p input{0}RI_{1}".format(os.sep, i)

    cmd = 'summarycalc {} {}'.format(summarycalc_switch, summarycalc_directory_switch)
    for summary in summaries:
        if 'id' in summary:
            summary_set = summary['id']
            cmd = '{0} -{1} fifo/{2}_S{1}_summary_P{3}'.format(cmd, summary_set, runtype, process_id)

    cmd = '{} < fifo/{}_P{} &'.format(cmd, runtype, process_id)
    print_command(filename, cmd)


def do_tees(runtype, analysis_settings, process_id, filename, process_counter):
    summaries = analysis_settings.get('{}_summaries'.format(runtype))
    if not summaries:
        return

    for summary in summaries:
        if 'id' in summary:
            process_counter['pid_monitor_count'] += 1
            summary_set = summary['id']
            cmd = 'tee < fifo/{}_S{}_summary_P{}'.format(runtype, summary_set, process_id)

            if summary.get('eltcalc'):
                cmd = '{} fifo/{}_S{}_summaryeltcalc_P{}'.format(cmd, runtype, summary_set, process_id)

            if summary.get('pltcalc'):
                cmd = '{} fifo/{}_S{}_summarypltcalc_P{}'.format(cmd, runtype, summary_set, process_id)

            if summary.get('summarycalc'):
                cmd = '{} fifo/{}_S{}_summarysummarycalc_P{}'.format(cmd, runtype, summary_set, process_id)

            if summary.get('aalcalc'):
                cmd = '{} work/{}_S{}_summaryaalcalc/P{}.bin'.format(cmd, runtype, summary_set, process_id)

            if summary.get('lec_output') and leccalc_enabled(summary['leccalc']):
                cmd = '{} work/{}_S{}_summaryleccalc/P{}.bin'.format(cmd, runtype, summary_set, process_id)

            cmd = '{} > /dev/null & pid{}=$!'.format(cmd, process_counter['pid_monitor_count'])
            print_command(filename, cmd)


def do_any(runtype, analysis_settings, process_id, filename, process_counter):
    summaries = analysis_settings.get('{}_summaries'.format(runtype))
    if not summaries:
        return

    for summary in summaries:
        if 'id' in summary:
            summary_set = summary['id']
            if summary.get('eltcalc'):
                cmd = 'eltcalc -s'
                if process_id == 1:
                    cmd = 'eltcalc'

                process_counter['pid_monitor_count'] += 1
                print_command(
                    filename,
                    "{3} < fifo/{0}_S{1}_summaryeltcalc_P{2} > work/kat/{0}_S{1}_eltcalc_P{2} & pid{4}=$!".format(
                        runtype, summary_set, process_id, cmd, process_counter['pid_monitor_count']
                    )
                )

            if summary.get("summarycalc"):
                cmd = 'summarycalctocsv -s'
                if process_id == 1:
                    cmd = 'summarycalctocsv'

                process_counter['pid_monitor_count'] += 1
                print_command(
                    filename,
                    '{3} < fifo/{0}_S{1}_summarysummarycalc_P{2} > work/kat/{0}_S{1}_summarycalc_P{2} & pid{4}=$!'.format(
                        runtype, summary_set, process_id, cmd, process_counter['pid_monitor_count']
                    )
                )

            if summary.get('pltcalc'):
                cmd = 'pltcalc -s'
                if process_id == 1:
                    cmd = 'pltcalc'

                process_counter['pid_monitor_count'] += 1
                print_command(
                    filename,
                    '{3} < fifo/{0}_S{1}_summarypltcalc_P{2} > work/kat/{0}_S{1}_pltcalc_P{2} & pid{4}=$!'.format(
                        runtype, summary_set, process_id, cmd, process_counter['pid_monitor_count']
                    )
                )


        print_command(filename, '')


def do_ri(analysis_settings, max_process_id, filename, process_counter, num_reinsurance_iterations):
    for process_id in range(1, max_process_id + 1):
        do_any(RUNTYPE_REINSURANCE_LOSS, analysis_settings, process_id, filename, process_counter)

    for process_id in range(1, max_process_id + 1):
        do_tees(RUNTYPE_REINSURANCE_LOSS, analysis_settings, process_id, filename, process_counter)

    for process_id in range(1, max_process_id + 1):
        do_summarycalcs(
            RUNTYPE_REINSURANCE_LOSS, analysis_settings, process_id, filename, num_reinsurance_iterations)


def do_il(analysis_settings, max_process_id, filename, process_counter):
    for process_id in range(1, max_process_id + 1):
        do_any(RUNTYPE_INSURED_LOSS, analysis_settings, process_id, filename, process_counter)

    for process_id in range(1, max_process_id + 1):
        do_tees(RUNTYPE_INSURED_LOSS, analysis_settings, process_id, filename, process_counter)

    for process_id in range(1, max_process_id + 1):
        do_summarycalcs(RUNTYPE_INSURED_LOSS, analysis_settings, process_id, filename)


def do_gul(analysis_settings, max_process_id, filename, process_counter):
    for process_id in range(1, max_process_id + 1):
        do_any(RUNTYPE_GROUNDUP_LOSS, analysis_settings, process_id, filename, process_counter)

    for process_id in range(1, max_process_id + 1):
        do_tees(RUNTYPE_GROUNDUP_LOSS, analysis_settings, process_id, filename, process_counter)

    for process_id in range(1, max_process_id + 1):
        do_summarycalcs(RUNTYPE_GROUNDUP_LOSS, analysis_settings, process_id, filename)


def do_il_make_fifo(analysis_settings, max_process_id, filename):
    for process_id in range(1, max_process_id + 1):
        do_make_fifos(RUNTYPE_INSURED_LOSS, analysis_settings, process_id, filename)


def do_gul_make_fifo(analysis_settings, max_process_id, filename):
    for process_id in range(1, max_process_id + 1):
        do_make_fifos(RUNTYPE_GROUNDUP_LOSS, analysis_settings, process_id, filename)


def do_ri_make_fifo(analysis_settings, max_process_id, filename):
    for process_id in range(1, max_process_id + 1):
        do_make_fifos(RUNTYPE_REINSURANCE_LOSS, analysis_settings, process_id, filename)


def do_il_remove_fifo(analysis_settings, max_process_id, filename):
    for process_id in range(1, max_process_id + 1):
        do_remove_fifos(RUNTYPE_INSURED_LOSS, analysis_settings, process_id, filename)


def do_gul_remove_fifo(analysis_settings, max_process_id, filename):
    for process_id in range(1, max_process_id + 1):
        do_remove_fifos(RUNTYPE_GROUNDUP_LOSS, analysis_settings, process_id, filename)


def do_ri_remove_fifo(analysis_settings, max_process_id, filename):
    for process_id in range(1, max_process_id + 1):
        do_remove_fifos(RUNTYPE_REINSURANCE_LOSS, analysis_settings, process_id, filename)


def do_waits(wait_variable, wait_count, filename):
    """
    Add waits to the script

    :param wait_variable: The type of wait
    :type wait_variable: str

    :param wait_count: The number of processes to wait for
    :type wait_count: int
    """
    if wait_count > 0:
        cmd = 'wait'
        for pid in range(1, wait_count + 1):
            cmd = '{} ${}{}'.format(cmd, wait_variable, pid)

        print_command(filename, cmd)
        print_command(filename, '')


def do_pwaits(filename, process_counter):
    """
    Add pwaits to the script
    """
    do_waits('pid', process_counter['pid_monitor_count'], filename)


def do_awaits(filename, process_counter):
    """
    Add awaits to the script
    """
    do_waits('apid', process_counter['apid_monitor_count'], filename)


def do_lwaits(filename, process_counter):
    """
    Add lwaits to the script
    """
    do_waits('lpid', process_counter['lpid_monitor_count'], filename)


def do_kwaits(filename, process_counter):
    """
    Add kwaits to the script
    """
    do_waits('kpid', process_counter['kpid_monitor_count'], filename)


def get_getmodel_cmd(number_of_samples, gul_threshold, use_random_number_file, coverage_output, item_output, **kwargs):
    """
    Gets the getmodel ktools command

    :param number_of_samples: The number of samples to run
    :type number_of_samples: int

    :param gul_threshold: The GUL threshold to use
    :type gul_threshold: float

    :param use_random_number_file: flag to use the random number file
    :type use_random_number_file: bool

    :param coverage_output: The coverage output
    :type coverage_output: str

    :param item_output: The item output
    :type item_output: str

    :return: The generated getmodel command
    """

    cmd = 'getmodel | gulcalc -S{} -L{}'.format(number_of_samples, gul_threshold)

    if use_random_number_file:
        cmd = '{} -r'.format(cmd)
    if coverage_output != '':
        cmd = '{} -c {}'.format(cmd, coverage_output)
    if item_output != '':
        cmd = '{} -i {}'.format(cmd, item_output)

    return cmd

def genbash(
    max_process_id, analysis_settings, filename, 
    num_reinsurance_iterations=0,
    _get_getmodel_cmd=get_getmodel_cmd, custom_args={}):
    """
    Generates a bash script containing ktools calculation instructions for an
    Oasis model.

    :param max_process_id: The number of processes to create
    :type max_process_id: int

    :param analysis_settings: The analysis settings
    :type analysis_settings: dict

    :param filename: The output file name
    :type filename: string

    :param num_reinsurance_iterations: The number of reinsurance iterations
    :type num_reinsurance_iterations: int
    
    :param get_getmodel_cmd: Method for getting the getmodel command, by default
        ``GenerateLossesCmd.get_getmodel_cmd`` is used.
    :type get_getmodel_cmd: callable
    """
    process_counter = Counter()

    use_random_number_file = False
    gul_output = False
    il_output = False
    ri_output = False

    # remove the file if it already exists
    if os.path.exists(filename):
        os.remove(filename)

    gul_threshold = analysis_settings.get('gul_threshold', 0)
    number_of_samples = analysis_settings.get('number_of_samples', 0)

    if 'model_settings' in analysis_settings and analysis_settings['model_settings'].get('use_random_number_file'):
        use_random_number_file = True

    if 'gul_output' in analysis_settings:
        gul_output = analysis_settings['gul_output']

    if 'il_output' in analysis_settings:
        il_output = analysis_settings['il_output']

    if 'ri_output' in analysis_settings:
        ri_output = analysis_settings['ri_output']

    print_command(filename, '#!/bin/bash')

    print_command(filename, '')

    print_command(filename, 'rm -R -f output/*')
    print_command(filename, 'rm -R -f fifo/*')
    print_command(filename, 'rm -R -f work/*')
    print_command(filename, '')

    print_command(filename, 'mkdir work/kat')

    if gul_output:
        do_gul_make_fifo(analysis_settings, max_process_id, filename)
        create_workfolders(RUNTYPE_GROUNDUP_LOSS, analysis_settings, filename)

    if il_output:
        do_il_make_fifo(analysis_settings, max_process_id, filename)
        create_workfolders(RUNTYPE_INSURED_LOSS, analysis_settings, filename)

    if ri_output:
        do_ri_make_fifo(analysis_settings, max_process_id, filename)
        create_workfolders(RUNTYPE_REINSURANCE_LOSS, analysis_settings, filename)
        print_command(filename, '')

    if ri_output:
        print_command(filename, '')
        print_command(filename, '# --- Do reinsurance loss computes ---')
        print_command(filename, '')
        do_ri(analysis_settings, max_process_id, filename, process_counter, num_reinsurance_iterations)

    if il_output:
        print_command(filename, '')
        print_command(filename, '# --- Do insured loss computes ---')
        print_command(filename, '')
        do_il(analysis_settings, max_process_id, filename, process_counter)

    if gul_output:
        print_command(filename, '')
        print_command(filename, '# --- Do ground up loss computes ---')
        print_command(filename, '')
        do_gul(analysis_settings, max_process_id, filename, process_counter)

    print_command(filename, '')
    
    for process_id in range(1, max_process_id + 1):

        ##! Should be able to streamline the logic a little
        if num_reinsurance_iterations > 0 and ri_output:
            
            getmodel_args = { 
                'number_of_samples'      : number_of_samples,
                'gul_threshold'          : gul_threshold,
                'use_random_number_file' : use_random_number_file,
                'coverage_output'        : 'fifo/gul_P{}'.format(process_id),
                'item_output'            : '-',
                'process_id'             : process_id,
                'max_process_id'         : max_process_id
            }
            getmodel_args.update(custom_args)
            getmodel_cmd = _get_getmodel_cmd(**getmodel_args)
            main_cmd = 'eve {0} {1} | {2} | fmcalc -a {3} | tee fifo/il_P{0}'.format(
                process_id, max_process_id, getmodel_cmd, 
                ALLOCATE_TO_ITEMS_BY_PREVIOUS_LEVEL_ALLOC_ID)
            for i in range(1, num_reinsurance_iterations + 1):
                main_cmd = "{0} | fmcalc -a {3} -n -p input{1}RI_{2}".format(
                    main_cmd, os.sep, i, ALLOCATE_TO_ITEMS_BY_PREVIOUS_LEVEL_ALLOC_ID)
            main_cmd = "{} > fifo/ri_P{} &".format(main_cmd, process_id)

            print_command(
                filename,
                main_cmd
            )

        elif gul_output and il_output:
            getmodel_args = { 
                'number_of_samples'      : number_of_samples,
                'gul_threshold'          : gul_threshold,
                'use_random_number_file' : use_random_number_file,
                'coverage_output'        : 'fifo/gul_P{}'.format(process_id),
                'item_output'            : '-',
                'process_id'             : process_id,
                'max_process_id'         : max_process_id
            }
            getmodel_args.update(custom_args)
            getmodel_cmd = _get_getmodel_cmd(**getmodel_args)
            main_cmd = 'eve {0} {1} | {2} | fmcalc -a {3} > fifo/il_P{0}  &'.format(
                process_id, max_process_id, getmodel_cmd,
                ALLOCATE_TO_ITEMS_BY_PREVIOUS_LEVEL_ALLOC_ID)

            print_command(
                filename,
                main_cmd
            )
        else:
            if gul_output and 'gul_summaries' in analysis_settings:
                getmodel_args = { 
                    'number_of_samples'      : number_of_samples,
                    'gul_threshold'          : gul_threshold,
                    'use_random_number_file' : use_random_number_file,
                    'coverage_output'        : '-',
                    'item_output'            : '',
                    'process_id'             : process_id,
                    'max_process_id'         : max_process_id
                }
                getmodel_args.update(custom_args)
                getmodel_cmd = _get_getmodel_cmd(**getmodel_args)
                print_command(
                    filename,
                    'eve {0} {1} | {2} > fifo/gul_P{0}  &'.format(process_id, max_process_id, getmodel_cmd)
                )
            if il_output and 'il_summaries' in analysis_settings:
                getmodel_args = { 
                    'number_of_samples'      : number_of_samples,
                    'gul_threshold'          : gul_threshold,
                    'use_random_number_file' : use_random_number_file,
                    'coverage_output'        : '',
                    'item_output'            : '-',
                    'process_id'             : process_id,
                    'max_process_id'         : max_process_id
                }
                getmodel_args.update(custom_args)
                getmodel_cmd = _get_getmodel_cmd(**getmodel_args)
                print_command(
                    filename,
                    "eve {0} {1} | {2} | fmcalc -a {3} > fifo/il_P{0}  &".format(
                        process_id, max_process_id, getmodel_cmd,
                        ALLOCATE_TO_ITEMS_BY_PREVIOUS_LEVEL_ALLOC_ID)
                )

    print_command(filename, '')

    do_pwaits(filename, process_counter)

    if ri_output:
        print_command(filename, '')
        print_command(filename, '# --- Do reinsurance loss kats ---')
        print_command(filename, '')
        do_kats(RUNTYPE_REINSURANCE_LOSS, analysis_settings, max_process_id, filename, process_counter)

    if il_output:
        print_command(filename, '')
        print_command(filename, '# --- Do insured loss kats ---')
        print_command(filename, '')
        do_kats(RUNTYPE_INSURED_LOSS, analysis_settings, max_process_id, filename, process_counter)

    if gul_output:
        print_command(filename, '')
        print_command(filename, '# --- Do ground up loss kats ---')
        print_command(filename, '')
        do_kats(RUNTYPE_GROUNDUP_LOSS, analysis_settings, max_process_id, filename, process_counter)

    do_kwaits(filename, process_counter)

    print_command(filename, '')
    do_post_wait_processing(RUNTYPE_REINSURANCE_LOSS, analysis_settings, filename, process_counter)
    do_post_wait_processing(RUNTYPE_INSURED_LOSS, analysis_settings, filename, process_counter)
    do_post_wait_processing(RUNTYPE_GROUNDUP_LOSS, analysis_settings, filename, process_counter)

    do_awaits(filename, process_counter)  # waits for aalcalc
    do_lwaits(filename, process_counter)  # waits for leccalc

    if gul_output:
        do_gul_remove_fifo(analysis_settings, max_process_id, filename)
        remove_workfolders(RUNTYPE_GROUNDUP_LOSS, analysis_settings, filename)

    print_command(filename, '')

    if ri_output:
        do_ri_remove_fifo(analysis_settings, max_process_id, filename)
        remove_workfolders(RUNTYPE_REINSURANCE_LOSS, analysis_settings, filename)

    if il_output:
        do_il_remove_fifo(analysis_settings, max_process_id, filename)
        remove_workfolders(RUNTYPE_INSURED_LOSS, analysis_settings, filename)
